# tests for NumPy __array_function__ support
import inspect
import re
import warnings
from importlib.metadata import version

import numpy as np
import pytest
from numpy.testing import assert_allclose
from packaging.version import Version

from unyt import A, K, Msun, cm, degC, delta_degC, dimensionless, g, km, rad, s
from unyt._array_functions import (
    _HANDLED_FUNCTIONS as HANDLED_FUNCTIONS,
    _UNSUPPORTED_FUNCTIONS as UNSUPPORTED_FUNCTIONS,
    _sanitize_bins,
    _sanitize_range,
)
from unyt.array import unyt_array, unyt_quantity
from unyt.exceptions import (
    UnitConversionError,
    UnitInconsistencyError,
    UnytError,
)
from unyt.testing import assert_array_equal_units
from unyt.unit_object import Unit
from unyt.unit_registry import UnitRegistry

NUMPY_VERSION = Version(version("numpy"))


VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD
POSITIONAL_ONLY = inspect.Parameter.POSITIONAL_ONLY
KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
POSITIONAL_OR_KEYWORD = inspect.Parameter.POSITIONAL_OR_KEYWORD

# this is a subset of NOT_HANDLED_FUNCTIONS for which there's nothing to do
# because they don't apply to (real) numeric types
# or they work as expected out of the box
# This is not necessarily complete !
NOOP_FUNCTIONS = {
    np.all,  # expects booleans
    np.amax,  # works out of the box (tested)
    np.amin,  # works out of the box (tested)
    np.angle,  # expects complex numbers
    np.any,  # works out of the box (tested)
    np.append,  # we get it for free with np.concatenate (tested)
    np.apply_along_axis,  # works out of the box (tested)
    np.argmax,  # returns pure numbers
    np.argmin,  # returns pure numbers
    np.argpartition,  # returns pure numbers
    np.argsort,  # returns pure numbers
    np.argwhere,  # returns pure numbers
    np.array_str,  # hooks into __str__
    np.atleast_1d,  # works out of the box (tested)
    np.atleast_2d,  # works out of the box (tested)
    np.atleast_3d,  # works out of the box (tested)
    np.can_cast,  # works out of the box (tested)
    np.common_type,  # works out of the box (tested)
    np.result_type,  # works out of the box (tested)
    np.iscomplex,  # works out of the box (tested)
    np.iscomplexobj,  # works out of the box (tested)
    np.isreal,  # works out of the box (tested)
    np.isrealobj,  # works out of the box (tested)
    np.nan_to_num,  # works out of the box (tested)
    np.nanargmax,  # return pure numbers
    np.nanargmin,  # return pure numbers
    np.nanmax,  # works out of the box (tested)
    np.nanmean,  # works out of the box (tested)
    np.nanmedian,  # works out of the box (tested)
    np.nanmin,  # works out of the box (tested)
    np.trim_zeros,  # works out of the box (tested)
    np.max,  # works out of the box (tested)
    np.mean,  # works out of the box (tested)
    np.median,  # works out of the box (tested)
    np.min,  # works out of the box (tested)
    np.ndim,  # return pure numbers
    np.shape,  # returns pure numbers
    np.size,  # returns pure numbers
    np.sort,  # works out of the box (tested)
    np.sum,  # works out of the box (tested)
    np.repeat,  # works out of the box (tested)
    np.tile,  # works out of the box (tested)
    np.shares_memory,  # works out of the box (tested)
    np.nonzero,  # works out of the box (tested)
    np.count_nonzero,  # returns pure numbers
    np.flatnonzero,  # works out of the box (tested)
    np.isneginf,  # works out of the box (tested)
    np.isposinf,  # works out of the box (tested)
    np.empty_like,  # works out of the box (tested)
    np.full_like,  # works out of the box (tested)
    np.ones_like,  # works out of the box (tested)
    np.zeros_like,  # works out of the box (tested)
    np.copy,  # works out of the box (tested)
    np.meshgrid,  # works out of the box (tested)
    np.transpose,  # works out of the box (tested)
    np.reshape,  # works out of the box (tested)
    np.resize,  # works out of the box (tested)
    np.roll,  # works out of the box (tested)
    np.rollaxis,  # works out of the box (tested)
    np.rot90,  # works out of the box (tested)
    np.expand_dims,  # works out of the box (tested)
    np.squeeze,  # works out of the box (tested)
    np.flip,  # works out of the box (tested)
    np.fliplr,  # works out of the box (tested)
    np.flipud,  # works out of the box (tested)
    np.delete,  # works out of the box (tested)
    np.partition,  # works out of the box (tested)
    np.broadcast_to,  # works out of the box (tested)
    np.broadcast_arrays,  # works out of the box (tested)
    np.split,  # works out of the box (tested)
    np.array_split,  # works out of the box (tested)
    np.dsplit,  # works out of the box (tested)
    np.hsplit,  # works out of the box (tested)
    np.vsplit,  # works out of the box (tested)
    np.swapaxes,  # works out of the box (tested)
    np.moveaxis,  # works out of the box (tested)
    np.nansum,  # works out of the box (tested)
    np.std,  # works out of the box (tested)
    np.nanstd,  # works out of the box (tested)
    np.nanvar,  # works out of the box (tested)
    np.nanprod,  # works out of the box (tested)
    np.diag,  # works out of the box (tested)
    np.diag_indices_from,  # returns pure numbers
    np.diagflat,  # works out of the box (tested)
    np.diagonal,  # works out of the box (tested)
    np.ravel,  # returns pure numbers
    np.ravel_multi_index,  # returns pure numbers
    np.unravel_index,  # returns pure numbers
    np.fix,  # works out of the box (tested)
    np.round,  # is implemented via np.around
    np.may_share_memory,  # returns pure numbers (booleans)
    np.linalg.matrix_power,  # works out of the box (tested)
    np.linalg.cholesky,  # works out of the box (tested)
    np.linalg.multi_dot,  # works out of the box (tested)
    np.linalg.matrix_rank,  # returns pure numbers
    np.linalg.qr,  # works out of the box (tested)
    np.linalg.slogdet,  # undefined units
    np.linalg.cond,  # works out of the box (tested)
    np.gradient,  # works out of the box (tested)
    np.cumsum,  # works out of the box (tested)
    np.nancumsum,  # works out of the box (tested)
    np.nancumprod,  # we get it for free with np.cumprod (tested)
    np.bincount,  # works out of the box (tested)
    np.unique,  # works out of the box (tested)
    np.min_scalar_type,  # returns dtypes
    np.extract,  # works out of the box (tested)
    np.setxor1d,  # we get it for free with previously implemented functions (tested)
    np.lexsort,  # returns pure numbers
    np.digitize,  # returns pure numbers
    np.tril_indices_from,  # returns pure numbers
    np.triu_indices_from,  # returns pure numbers
    np.imag,  # works out of the box (tested)
    np.real,  # works out of the box (tested)
    np.real_if_close,  # works out of the box (tested)
    np.einsum_path,  # returns pure numbers
    np.cov,  # returns pure numbers
    np.corrcoef,  # returns pure numbers
    np.compress,  # works out of the box (tested)
    np.take_along_axis,  # works out of the box (tested)
}


if NUMPY_VERSION >= Version("2.0.0dev0"):
    # the following all work out of the box (tested)
    NOOP_FUNCTIONS |= {
        np.linalg.cross,
        np.linalg.diagonal,
        np.linalg.matmul,
        np.linalg.matrix_norm,
        np.linalg.matrix_transpose,
        np.linalg.svdvals,
        np.linalg.tensordot,
        np.linalg.trace,
        np.linalg.vecdot,
        np.linalg.vector_norm,
        np.astype,
        np.matrix_transpose,
        np.unique_all,
        np.unique_counts,
        np.unique_inverse,
        np.unique_values,
    }

if NUMPY_VERSION >= Version("2.1.0dev0"):
    NOOP_FUNCTIONS |= {
        np.unstack,
        np.cumulative_sum,
    }

if NUMPY_VERSION >= Version("2.4.1"):
    NOOP_FUNCTIONS |= {np.average}

# Functions for which behaviour is intentionally left to default
IGNORED_FUNCTIONS = {
    np.i0,
    # IO functions (no way to add units)
    np.save,
    np.savez,
    np.savez_compressed,
}

# map subsets of deprecated functions to the version they were removed
# so that we can drop the ones that are not present in any version we support
DEPRECATED_FUNCTIONS = {
    Version("1.20"): {
        "fv",
        "ipmt",
        "irr",
        "mirr",
        "nper",
        "npv",
        "pmt",
        "ppmt",
        "pv",
        "rate",
    },
    Version("1.22"): {
        "alen",
        "asscalar",
    },
    Version("2.0.0b1"): {
        "msort",
        "product",
        "cumproduct",
        "round_",
        "sometrue",
        "alltrue",
    },
    # functions that are deprecated but not yet removed in any known version
    # should be added here
    Version("999.999.999"): set(),
}

NOT_HANDLED_FUNCTIONS = NOOP_FUNCTIONS | UNSUPPORTED_FUNCTIONS | IGNORED_FUNCTIONS

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    for removal_version, functions in DEPRECATED_FUNCTIONS.items():
        if NUMPY_VERSION >= removal_version:
            continue
        for func in functions:
            NOT_HANDLED_FUNCTIONS.add(getattr(np, func))


def get_decorators(func):
    # adapted from
    # https://stackoverflow.com/questions/3232024/introspection-to-get-decorator-names-on-a-method
    import ast
    import inspect

    target = func
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = []
        for n in node.decorator_list:
            name = ""
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[node.name].append(name)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    try:
        node_iter.visit(ast.parse(inspect.getsource(target)))
        return decorators[func.__name__]
    except TypeError:
        # may be raised if inspecting a C compiled function
        # in which case, we return with empty hands
        return []


def get_wrapped_functions(*modules):
    """get functions that support __array_function__ in modules

    This was adapted from astropy's tests
    """
    wrapped_functions = {}
    for mod in modules:
        for name, f in mod.__dict__.items():
            if callable(f) and hasattr(f, "__wrapped__"):
                if (
                    f is np.printoptions
                    or f.__name__.startswith("_")
                    or "deprecate" in get_decorators(f)
                ):
                    continue
                wrapped_functions[mod.__name__ + "." + name] = f
    return dict(sorted(wrapped_functions.items()))


def test_wrapping_completeness():
    """Ensure we wrap all numpy functions that support __array_function__"""
    handled_numpy_functions = set(HANDLED_FUNCTIONS.keys())
    # ensure no functions appear in both NOT_HANDLED_FUNCTIONS and HANDLED_FUNCTIONS
    assert NOT_HANDLED_FUNCTIONS.isdisjoint(handled_numpy_functions), (
        NOT_HANDLED_FUNCTIONS.intersection(handled_numpy_functions)
    )
    # get list of functions that support wrapping by introspection on numpy module
    wrappable_functions = get_wrapped_functions(np, np.fft, np.linalg)
    for function in HANDLED_FUNCTIONS:
        # ensure we only have wrappers for functions that support wrapping
        assert function in wrappable_functions.values()
    all_funcs = NOT_HANDLED_FUNCTIONS.union(handled_numpy_functions)
    # ensure all functions in numpy that support wrapping either have wrappers
    # or are explicitly whitelisted
    for function in wrappable_functions.values():
        assert function in all_funcs


@pytest.mark.parametrize(
    "arrays",
    [
        [np.array([1]), [2] * Unit()],
        [np.array([1]), [2] * Unit(registry=UnitRegistry())],
        [[1], [2] * Unit()],
    ],
)
def test_unit_validation(arrays):
    # see https://github.com/yt-project/unyt/issues/462
    # numpy.concatenate isn't essential to this test
    # what we're really testing is the unit consistency validation
    # underneath, but we do so using public API
    res = np.concatenate(arrays)
    assert res.units.is_dimensionless


def test_unit_validation_dimensionless_factors():
    # see https://github.com/yt-project/unyt/issues/477
    # numpy.concatenate isn't essential to this test
    # what we're really testing is the unit consistency validation
    # underneath, but we do so using public API
    res = np.concatenate([[1] * cm, unyt_array([1], "cm*dimensionless")])
    assert res.units is cm


def test_array_repr():
    arr = [1, 2, 3] * cm
    assert re.fullmatch(r"unyt_array\(\[1, 2, 3\], (units=)?'cm'\)", np.array_repr(arr))


def test_dot_vectors():
    a = [1, 2, 3] * cm
    b = [1, 2, 3] * s
    res = np.dot(a, b)
    assert res.units == cm * s
    assert res.d == 14


# NOTE: explicitly setting the dtype of out arrays as `dtype=np.int_` is the
# cross-platform way to guarantee that their dtype matches that of np.arange(x)
# (on POSIX system it's int64, while on windows it's int32)
@pytest.mark.parametrize(
    "out",
    [
        None,
        np.empty((3, 3), dtype=np.int_),
        np.empty((3, 3), dtype=np.int_, order="C") * cm * s,
        np.empty((3, 3), dtype=np.int_, order="C") * km * s,
    ],
    ids=[
        "None",
        "pure ndarray",
        "same units",
        "convertible units",
    ],
)
def test_dot_matrices(out):
    a = np.reshape(np.arange(9), (3, 3)) * cm
    b = np.reshape(np.arange(9), (3, 3)) * s

    res = np.dot(a, b, out=out)

    if out is not None:
        np.testing.assert_array_equal(res, out)
        assert np.shares_memory(res, out)
        assert isinstance(res, unyt_array)
        assert isinstance(out, np.ndarray)

    if isinstance(out, unyt_array):
        # check that the result can be converted to predictable units
        res.in_units("cm * s")
        assert out.units == res.units


def test_dot_mixed_ndarray_unyt_array():
    a = np.ones((3, 3))
    b = np.ones((3, 3)) * cm

    res = np.dot(a, b)

    assert isinstance(res, unyt_array)
    assert res.units == cm

    out = np.zeros((3, 3))
    res = np.dot(a, b, out=out)

    assert isinstance(res, unyt_array)
    assert type(out) is np.ndarray
    assert np.shares_memory(out, res)
    np.testing.assert_array_equal(out, res)

    out = np.zeros((3, 3)) * km
    res = np.dot(a, b, out=out)

    assert isinstance(res, unyt_array)
    assert isinstance(out, unyt_array)
    assert res.units == out.units == cm
    assert np.shares_memory(res, out)

    # check this works with an ndarray as the first operand
    out = np.zeros((3, 3)) * km
    res = np.dot(b, a, out=out)

    assert isinstance(res, unyt_array)
    assert isinstance(out, unyt_array)
    assert res.units == out.units == cm
    assert np.shares_memory(res, out)


def test_invalid_dot_matrices():
    a = np.reshape(np.arange(9), (3, 3)) * cm
    b = np.reshape(np.arange(9), (3, 3)) * s

    out = np.empty((3, 3), dtype=np.int_, order="C") * s**2
    res = np.dot(a, b, out=out)

    np.testing.assert_array_equal(res, out)
    assert out.units == res.units == cm * s


def test_vdot():
    a = np.arange(9) * cm
    b = np.arange(9) * s
    res = np.vdot(a, b)
    assert res.units == cm * s


def test_inner():
    a = np.array([1, 2, 3]) * cm
    b = np.array([0, 1, 0]) * s
    res = np.inner(a, b)
    assert res.d == 2
    assert res.units == cm * s


def test_outer():
    a = np.array([1, 2, 3]) * cm
    b = np.array([0, 1, 0]) * s
    res = np.outer(a, b)
    expected = np.array([[0, 1, 0], [0, 2, 0], [0, 3, 0]])
    np.testing.assert_array_equal(res.ndview, expected)
    assert res.units == cm * s


def test_kron():
    a = np.eye(2) * cm
    b = np.ones((2, 2)) * s
    res = np.kron(a, b)
    assert res.units == cm * s


def test_linalg_inv():
    rng = np.random.default_rng(0)
    arr = rng.random((3, 3)) * cm
    iarr = np.linalg.inv(arr)
    assert 1 * iarr.units == 1 / cm


def test_linalg_tensorinv():
    a = np.reshape(np.eye(4 * 6), (4, 6, 8, 3)) * cm
    ia = np.linalg.tensorinv(a)
    assert 1 * ia.units == 1 / cm


def test_linalg_pinv():
    rng = np.random.default_rng(0)
    a = rng.standard_normal(size=(9, 6)) * cm
    B = np.linalg.pinv(a)
    assert 1 * B.units == 1 / cm
    np.testing.assert_allclose(a, np.dot(a, np.dot(B, a)))
    np.testing.assert_allclose(B, np.dot(B, np.dot(a, B)))


# see https://github.com/numpy/numpy/issues/22444
@pytest.mark.xfail(
    reason=(
        "as of numpy 1.21.2, the __array_function__ protocol doesn't let "
        "us overload np.linalg.pinv for stacks of matrices"
    )
)
def test_matrix_stack_linalg_pinv():
    stack = [np.eye(4) * g for _ in range(3)]
    B = np.linalg.pinv(stack)
    assert 1 * B.units == 1 / g


# see https://github.com/numpy/numpy/issues/22444
@pytest.mark.xfail(
    reason=(
        "as of numpy 1.21.2, the __array_function__ protocol doesn't let "
        "us overload np.linalg.pinv for stacks of matrices"
    )
)
def test_invalid_matrix_stack_linalg_pinv():
    stack = [np.eye(4) * g, np.eye(4) * s]
    with pytest.raises(
        TypeError,
        match=re.escape(
            "numpy.linalg.pinv cannot operate on a stack "
            "of matrices with different units."
        ),
    ):
        np.linalg.pinv(stack)


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.diagonal is new in numpy 2.0"
)
def test_linalg_diagonal():
    a = np.eye(3) * cm
    b = np.linalg.diagonal(a)
    assert type(b) is unyt_array
    assert b.units == a.units


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.trace is new in numpy 2.0"
)
def test_linalg_trace():
    a = np.eye(3) * cm
    b = np.linalg.trace(a)
    assert type(b) is unyt_quantity
    assert b.units == a.units


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.outer is new in numpy 2.0"
)
def test_linalg_outer():
    a = np.arange(10) * cm
    assert_array_equal_units(np.linalg.outer(a, a), np.outer(a, a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.cross is new in numpy 2.0"
)
def test_linalg_cross():
    a = np.arange(3) * cm
    assert_array_equal_units(np.linalg.cross(a, a), np.cross(a, a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.matmul is new in numpy 2.0"
)
def test_linalg_matmul():
    a = np.eye(3) * cm
    assert_array_equal_units(np.linalg.matmul(a, a), np.matmul(a, a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"),
    reason="linalg.matrix_norm is new in numpy 2.0",
)
def test_linalg_matrix_norm():
    a = np.eye(3) * cm
    assert_array_equal_units(np.linalg.matrix_norm(a), np.linalg.norm(a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="matrix_transpose is new in numpy 2.0"
)
@pytest.mark.parametrize("namespace", [None, "linalg"])
def test_matrix_transpose(namespace):
    if namespace is None:
        func = np.matrix_transpose
    else:
        func = getattr(np, namespace).matrix_transpose
    a = np.arange(0, 9).reshape(3, 3)
    assert_array_equal_units(func(a), np.transpose(a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.vecdot is new in numpy 2.0"
)
def test_linalg_vecdot():
    a = np.arange(0, 9)
    assert_array_equal_units(np.linalg.vecdot(a, a), np.vdot(a, a))


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"),
    reason="linalg.vector_norm is new in numpy 2.0",
)
def test_linalg_vector_norm():
    a = np.arange(0, 9)
    assert_array_equal_units(np.linalg.vector_norm(a), np.linalg.norm(a))


def test_linalg_svd():
    rng = np.random.default_rng(0)
    a = (rng.standard_normal(size=(9, 6)) + 1j * rng.standard_normal(size=(9, 6))) * cm
    u, s, vh = np.linalg.svd(a)
    assert type(u) is np.ndarray
    assert type(vh) is np.ndarray
    assert type(s) is unyt_array
    assert s.units == cm

    s = np.linalg.svd(a, compute_uv=False)
    assert type(s) is unyt_array
    assert s.units == cm


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.svdvals is new in numpy 2.0"
)
def test_linalg_svdvals():
    q = np.arange(9).reshape(3, 3) * cm

    _, ref, _ = np.linalg.svd(q)
    res = np.linalg.svdvals(q)
    assert type(res) is unyt_array
    assert_allclose(res, ref, rtol=5e-16)


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="linalg.tensordot is new in numpy 2.0"
)
def test_linalg_tensordot():
    q = np.arange(9).reshape(3, 3) * cm
    ref = np.tensordot(q, q)
    res = np.linalg.tensordot(q, q)
    assert_array_equal_units(res, ref)


class TestHistograms:
    def test_sanitize_range_on_None(self):
        assert _sanitize_range(None, [cm]) is None

    def test_sanitize_range_units_match(self):
        input_range = (1 * cm, 2 * cm)
        expected_range = (1 * cm, 2 * cm)
        assert_array_equal_units(_sanitize_range(input_range, [cm]), expected_range)

    def test_sanitize_range_units_mismatch(self):
        input_range = (1 * cm, 2)
        with pytest.raises(
            TypeError, match="Elements of range must both have a 'units' attribute."
        ):
            _sanitize_range(input_range, [cm, g])

    def test_sanitize_range_pure_scalars(self):
        input_range = (1, 2)
        expected_range = (1 * cm, 2 * cm)
        assert_array_equal_units(_sanitize_range(input_range, [cm]), expected_range)

    def test_sanitize_bins_integer(self):
        a = np.arange(10) * cm
        bins = 10
        assert _sanitize_bins(a, bins) == bins

    def test_sanitize_bins_convertible(self):
        a = np.arange(10) * cm
        bins = np.array([0, 1, 2]) * km
        np.testing.assert_array_equal(_sanitize_bins(a, bins), bins.to_value(a.units))

    def test_sanitize_bins_missing_units_but_dimensionless(self):
        a = np.arange(10) * dimensionless
        bins = np.array([0, 1, 2])
        np.testing.assert_array_equal(_sanitize_bins(a, bins), bins)

    def test_sanitize_bins_without_units(self):
        a = np.arange(10)
        bins = np.array([0, 1, 2])
        np.testing.assert_array_equal(_sanitize_bins(a, bins), bins)

    def test_histogram(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        counts, bins = np.histogram(arr, bins=10, range=(arr.min(), arr.max()))
        assert type(counts) is np.ndarray
        assert bins.units == arr.units

    def test_histogram_implicit_units(self):
        # see https://github.com/yt-project/unyt/issues/465
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        counts, bins = np.histogram(
            arr, bins=10, range=(arr.min().value, arr.max().value)
        )
        assert type(counts) is np.ndarray
        assert bins.units == arr.units

    def test_histogram_with_density(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        density, bins = np.histogram(
            arr, bins=10, range=(arr.min(), arr.max()), density=True
        )
        assert type(density) is unyt_array
        assert density.units == arr.units**-1
        assert bins.units == arr.units

    def test_histogram_with_weights(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        w = rng.uniform(size=1000) * g
        wcounts, wbins = np.histogram(
            arr, bins=10, range=(arr.min(), arr.max()), weights=w
        )
        assert type(wcounts) is unyt_array
        assert wcounts.units == w.units
        assert wbins.units == arr.units

    def test_histogram_with_weights_and_density(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        w = rng.uniform(size=1000) * g
        wdensity, wdbins = np.histogram(
            arr, bins=10, range=(arr.min(), arr.max()), density=True, weights=w
        )
        assert type(wdensity) is unyt_array
        assert wdensity.units == w.units / arr.units
        assert wdbins.units == arr.units

    def test_histogram_with_weights_and_dimless_arr(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        w = rng.uniform(size=1000) * g
        wcounts2, wbins2 = np.histogram(arr.to_value(arr.units), weights=w)
        assert type(wcounts2) is unyt_array
        assert wcounts2.units == w.units
        assert not hasattr(wbins2, "units")

    def test_histogram2d(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        counts, xbins, ybins = np.histogram2d(x, y)
        assert counts.ndim == 2
        assert xbins.units == x.units
        assert ybins.units == y.units

    def test_histogram2d_with_density(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        density, xbins, ybins = np.histogram2d(x, y, density=True)
        assert density.ndim == 2
        assert type(density) is unyt_array
        assert density.units == (x.units * y.units) ** -1
        assert xbins.units == x.units
        assert ybins.units == y.units

    def test_histogram2d_with_weights(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        w = rng.uniform(size=100) * g
        wcounts, xwbins, ywbins = np.histogram2d(x, y, weights=w)
        assert wcounts.ndim == 2
        assert type(wcounts) is unyt_array
        assert wcounts.units == w.units
        assert xwbins.units == x.units
        assert ywbins.units == y.units

    def test_histogram2d_with_weights_and_density(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        w = rng.uniform(size=100) * g
        wdensity, xwdbins, ywdbins = np.histogram2d(x, y, weights=w, density=True)
        assert wdensity.ndim == 2
        assert type(wdensity) is unyt_array
        assert wdensity.units == w.units / (x.units * y.units)
        assert xwdbins.units == x.units
        assert ywdbins.units == y.units

    def test_histogram2d_with_weights_and_dimless_arr(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        w = rng.uniform(size=100) * g
        wcounts2, xwbins2, ywbins2 = np.histogram2d(
            x.to_value(x.units), y.to_value(y.units), weights=w
        )
        assert type(wcounts2) is unyt_array
        assert wcounts2.units == w.units
        assert not hasattr(xwbins2, "units")
        assert not hasattr(ywbins2, "units")

    def test_histogramdd(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(size=100) * s
        z = rng.normal(size=100) * g
        counts, (xbins, ybins, zbins) = np.histogramdd((x, y, z))
        assert counts.ndim == 3
        assert xbins.units == x.units
        assert ybins.units == y.units
        assert zbins.units == z.units

    def test_histogramdd_with_density(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        z = rng.normal(size=100) * g
        density, (xbins, ybins, zbins) = np.histogramdd((x, y, z), density=True)
        assert density.ndim == 3
        assert type(density) is unyt_array
        assert density.units == (x.units * y.units * z.units) ** -1
        assert xbins.units == x.units
        assert ybins.units == y.units
        assert zbins.units == z.units

    def test_histogramdd_with_weights(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        z = rng.normal(size=100) * g
        w = rng.uniform(size=100) * K
        wcounts, (xwbins, ywbins, zwbins) = np.histogramdd((x, y, z), weights=w)
        assert wcounts.ndim == 3
        assert type(wcounts) is unyt_array
        assert wcounts.units == w.units
        assert xwbins.units == x.units
        assert ywbins.units == y.units
        assert zwbins.units == z.units

    def test_histogramdd_with_weights_and_density(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        z = rng.normal(size=100) * g
        w = rng.uniform(size=100) * K
        wdensity, (xwdbins, ywdbins, zwdbins) = np.histogramdd(
            (x, y, z), weights=w, density=True
        )
        assert wdensity.ndim == 3
        assert type(wdensity) is unyt_array
        assert wdensity.units == w.units / (x.units * y.units * z.units)
        assert xwdbins.units == x.units
        assert ywdbins.units == y.units
        assert zwdbins.units == z.units

    def test_histogramdd_with_weights_and_dimless_arr(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100) * s
        z = rng.normal(size=100) * g
        w = rng.uniform(size=100) * K
        wcounts2, (xwbins2, ywbins2, zwbins2) = np.histogramdd(
            (x.to_value(x.units), y.to_value(y.units), z.to_value(z.units)), weights=w
        )
        assert type(wcounts2) is unyt_array
        assert wcounts2.units == w.units
        assert not hasattr(xwbins2, "units")
        assert not hasattr(ywbins2, "units")
        assert not hasattr(zwbins2, "units")

    @pytest.mark.parametrize("weights", [None, [0, 1, 2], [0, 1, 2] * cm])
    def test_histogramdd_recursion(self, weights):
        # regression test for https://github.com/yt-project/unyt/issues/540
        sample = [unyt_array(np.arange(3), Msun)]
        np.histogramdd(sample, density=True, weights=weights)

    def test_histogramdd_with_density_and_non_unyt_samples(self):
        rng = np.random.default_rng(0)
        x = rng.normal(size=100) * cm
        y = rng.normal(loc=10, size=100)
        z = rng.normal(size=100) * g
        counts, (xbins, ybins, zbins) = np.histogramdd((x, y, z), density=True)
        assert counts.units == (x.units * z.units) ** -1

    def test_histogram_bin_edges(self):
        rng = np.random.default_rng(0)
        arr = rng.normal(size=1000) * cm
        bins = np.histogram_bin_edges(arr)
        assert type(bins) is unyt_array
        assert bins.units == arr.units

    def test_histogram_input_and_bin_units_mismatch(self):
        # regression test for https://github.com/yt-project/unyt/issues/609
        rng = np.random.default_rng(0)
        arr = rng.uniform(size=10) * cm
        bins = np.linspace(0, 1, 10) * km
        hist, resulting_bins = np.histogram(arr, bins=bins)
        assert np.all(bins == resulting_bins)

    def test_histogram_input_and_bin_dimensions_mismatch(self):
        # regression test for https://github.com/yt-project/unyt/issues/609
        rng = np.random.default_rng(0)
        arr = rng.uniform(size=10) * cm
        bins = np.linspace(0, 1, 10) * s
        with pytest.raises(UnitConversionError):
            np.histogram(arr, bins=bins)


def test_concatenate():
    rng = np.random.default_rng(0)
    x1 = rng.normal(size=100) * cm
    x2 = rng.normal(size=100) * cm
    res = np.concatenate((x1, x2))
    assert res.units == cm
    assert res.shape == (200,)


def test_concatenate_different_units():
    rng = np.random.default_rng(0)
    x1 = rng.normal(size=100) * cm
    x2 = rng.normal(size=100) * s
    with pytest.raises(
        UnitInconsistencyError,
        match=(
            r"Expected all unyt_array arguments to have identical units\. "
            r"Received mixed units \(cm, s\)"
        ),
    ):
        np.concatenate((x1, x2))


def test_cross():
    x1 = [1, 2, 3] * cm
    x2 = [4, 5, 6] * s
    res = np.cross(x1, x2)
    assert res.units == cm * s


def test_intersect1d():
    x1 = [1, 2, 3, 4, 5, 6, 7, 8] * cm
    x2 = [0, 2, 4, 6, 8] * cm
    res = np.intersect1d(x1, x2)
    assert res.units == cm
    np.testing.assert_array_equal(res, [2, 4, 6, 8])


def test_intersect1d_return_indices():
    x1 = [1, 2, 3, 4, 5, 6, 7, 8] * cm
    x2 = [0, 2, 4, 6, 8] * cm
    ures = np.intersect1d(x1, x2, return_indices=True)
    rres = np.intersect1d(x1.d, x2.d, return_indices=True)
    np.testing.assert_array_equal(ures, rres)


def test_union1d():
    x1 = [-1, 0, 1] * cm
    x2 = [-2, -1, -3] * cm
    res = np.union1d(x1, x2)
    assert res.units == cm
    np.testing.assert_array_equal(res, [-3, -2, -1, 0, 1])


def test_linalg_norm():
    x = [1, 1, 1] * s
    res = np.linalg.norm(x)
    assert res.units == s
    assert res == pytest.approx(np.sqrt(3))


@pytest.mark.parametrize("func", [np.vstack, np.hstack, np.dstack, np.column_stack])
def test_xstack(func):
    x1 = [0, 1, 2] * cm
    x2 = [3, 4, 5] * cm
    res = func((x1, x2))
    assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.parametrize(
    "axis, expected", [(0, [[0, 1, 2], [3, 4, 5]]), (1, [[0, 3], [1, 4], [2, 5]])]
)
def test_stack(axis, expected):
    x1 = [0, 1, 2] * cm
    x2 = [3, 4, 5] * cm
    res = np.stack((x1, x2), axis=axis)
    assert res.units == cm
    np.testing.assert_array_equal(res, expected)


def test_amax():
    x1 = [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]] * cm
    res = np.amax(x1)
    assert type(res) is unyt_quantity
    res = np.amax(x1, axis=1)
    assert type(res) is unyt_array


def test_amin():
    x1 = [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]] * cm
    res = np.amin(x1)
    assert type(res) is unyt_quantity
    res = np.amin(x1, axis=1)
    assert type(res) is unyt_array


def test_around():
    x1 = [[1, 2, 3], [1, 2, 3], [1, 2, 3.0]] * g
    res = np.around(x1, 2)
    assert type(res) is unyt_array
    assert res.units == g


def test_atleast_nd():
    x0 = 1.0 * cm

    x1 = np.atleast_1d(x0)
    assert type(x1) is unyt_array
    assert x1.ndim == 1
    assert x1.units == cm

    x2 = np.atleast_2d(x0)
    assert type(x2) is unyt_array
    assert x2.ndim == 2
    assert x2.units == cm

    x3 = np.atleast_3d(x0)
    assert type(x3) is unyt_array
    assert x3.ndim == 3
    assert x3.units == cm


def test_average():
    x1 = [0.0, 1.0, 2.0] * cm
    res = np.average(x1)
    assert type(res) is unyt_quantity
    assert res == 1 * cm


def test_average_with_returned():
    # regression test for https://github.com/yt-project/unyt/issues/608
    x = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]] * cm
    w = [1.0, 2.0, 3.0] * g
    avg, wsum = np.average(x, weights=w, axis=0, returned=True)
    assert type(wsum) is unyt_array
    assert wsum.units == g


def test_trim_zeros():
    x1 = [0, 1, 2, 3, 0] * cm
    res = np.trim_zeros(x1)
    assert type(res) is unyt_array


def test_any():
    assert not np.any([0, 0, 0] * cm)
    assert np.any([1, 0, 0] * cm)

    x = [1, 2, 3] * cm
    assert np.any(x >= 3)
    assert np.any(x >= 3 * cm)
    assert not np.any(x >= 3 * km)


def test_append():
    a = [0, 1, 2, 3] * cm
    b = np.append(a, [4, 5, 6] * cm)
    assert type(b) is unyt_array
    assert b.units == cm


def test_append_inconsistent_units():
    a = [0, 1, 2, 3] * cm
    with pytest.raises(
        UnitInconsistencyError,
        match=re.escape(
            r"Expected all unyt_array arguments to have identical units. "
            r"Received mixed units (cm, dimensionless)"
        ),
    ):
        np.append(a, [4, 5, 6])


@pytest.mark.skipif(
    NUMPY_VERSION >= Version("2.0.0dev0"), reason="np.asfarray is removed in numpy 2.0"
)
def test_asfarray():
    x1 = np.eye(3, dtype="int64") * cm

    x2 = np.asfarray(x1)  # noqa: NPY201
    assert type(x2) is unyt_array
    assert x2.units == cm
    assert x2.dtype == "float64"


def test_block():
    x1 = 1 * np.ones((3, 3)) * cm
    x2 = 2 * np.ones((3, 1)) * cm
    res = np.block([[x1, x2]])
    assert type(res) is unyt_array
    assert res.units == cm


def test_block_units_inconsistency():
    # check that unit inconsistency is correctly detected
    # for nested lists
    x1 = 1 * np.ones((3, 3)) * cm
    x2 = [3 * cm, 3 * cm, 3 * km]
    with pytest.raises(UnitInconsistencyError):
        np.block([[x1, x2]])


def test_can_cast():
    a = [0, 1, 2] * cm
    assert np.can_cast(a, "float64")
    assert np.can_cast(a, "int64")
    assert not np.can_cast(a, "float16")


def test_isreal_like():
    a = [1, 2, 3] * cm
    assert np.all(np.isreal(a))
    assert np.isrealobj(a)
    assert not np.any(np.iscomplex(a))
    assert not np.iscomplexobj(a)

    b = [1j, 2j, 3j] * cm
    assert not np.any(np.isreal(b))
    assert not np.isrealobj(b)
    assert np.all(np.iscomplex(b))
    assert np.iscomplexobj(b)


@pytest.mark.parametrize(
    "func",
    [
        np.fft.fft,
        np.fft.hfft,
        np.fft.rfft,
        np.fft.ifft,
        np.fft.ihfft,
        np.fft.irfft,
    ],
)
def test_fft_1D(func):
    x1 = [0, 1, 2] * cm
    res = func(x1)
    assert type(res) is unyt_array
    assert res.units == (1 * cm).units


@pytest.mark.parametrize(
    "func",
    [
        np.fft.fft2,
        np.fft.fftn,
        np.fft.rfft2,
        np.fft.rfftn,
        np.fft.ifft2,
        np.fft.ifftn,
        np.fft.irfft2,
        np.fft.irfftn,
    ],
)
def test_fft_ND(func):
    x1 = [[0, 1, 2], [0, 1, 2], [0, 1, 2]] * cm
    res = func(x1)
    assert type(res) is unyt_array
    assert res.units == (1 * cm).units


@pytest.mark.parametrize("func", [np.fft.fftshift, np.fft.ifftshift])
def test_fft_shift(func):
    x1 = [[0, 1, 2], [0, 1, 2], [0, 1, 2]] * cm
    res = func(x1)
    assert type(res) is unyt_array
    assert res.units == cm


if NUMPY_VERSION >= Version("2.0.0dev0"):
    _trapezoid_func = np.trapezoid
else:
    _trapezoid_func = np.trapz  # noqa: NPY201


def test_trapezoid_no_x():
    y = [0, 1, 2, 3] * cm
    res = _trapezoid_func(y)
    assert type(res) is unyt_quantity
    assert res.units == cm


def test_trapezoid_with_raw_x():
    y = [0, 1, 2, 3] * cm
    x = [0, 1, 2, 3]
    res = _trapezoid_func(y, x)
    assert type(res) is unyt_quantity
    assert res.units == cm


def test_trapezoid_with_unit_x():
    y = [0, 1, 2, 3] * cm
    x = [0, 1, 2, 3] * s
    res = _trapezoid_func(y, x)
    assert type(res) is unyt_quantity
    assert res.units == cm * s


def test_trapezoid_with_raw_dx():
    y = [0, 1, 2, 3] * cm
    dx = 2.0
    res = _trapezoid_func(y, dx=dx)
    assert type(res) is unyt_quantity
    assert res.units == cm


def test_trapezoid_with_unit_dx():
    y = [0, 1, 2, 3] * cm
    dx = 2.0 * s
    res = _trapezoid_func(y, dx=dx)
    assert type(res) is unyt_quantity
    assert res.units == cm * s


@pytest.mark.parametrize(
    "op",
    ["min", "max", "mean", "median", "sum", "nanmin", "nanmax", "nanmean", "nanmedian"],
)
def test_scalar_reduction(op):
    x = [0, 1, 2] * cm
    res = getattr(np, op)(x)
    assert type(res) is unyt_quantity
    assert res.units == cm


@pytest.mark.parametrize("op", ["sort", "sort_complex"])
def test_sort(op):
    x = [2, 0, 1] * cm
    res = getattr(np, op)(x)
    assert type(res) is unyt_array
    assert res.units == cm


def test_repeat():
    x = [2, 0, 1] * cm
    res = np.repeat(x, 2)
    assert type(res) is unyt_array
    assert res.units == cm


def test_tile():
    x = [2, 0, 1] * cm
    res = np.tile(x, (2, 3))
    assert type(res) is unyt_array
    assert res.units == cm


def test_shares_memory():
    x = [1, 2, 3] * cm
    assert np.shares_memory(x, x.view(np.ndarray))


def test_nonzero():
    x = [1, 2, 0] * cm
    res = np.nonzero(x)
    assert len(res) == 1
    np.testing.assert_array_equal(res[0], [0, 1])

    res2 = np.flatnonzero(x)
    np.testing.assert_array_equal(res[0], res2)


def test_isinf():
    x = [1, float("inf"), float("-inf")] * cm
    res = np.isneginf(x)
    np.testing.assert_array_equal(res, [False, False, True])
    res = np.isposinf(x)
    np.testing.assert_array_equal(res, [False, True, False])


def test_allclose():
    x = [1, 2, 3] * cm
    y = [1, 2, 3] * km
    assert not np.allclose(x, y)


@pytest.mark.parametrize(
    "a, b, expected",
    [
        ([1, 2, 3] * cm, [1, 2, 3] * km, [False] * 3),
        ([1, 2, 3] * cm, [1, 2, 3], [True] * 3),
        ([1, 2, 3] * K, [-272.15, -271.15, -270.15] * degC, [True] * 3),
    ],
)
def test_isclose(a, b, expected):
    res = np.isclose(a, b)
    np.testing.assert_array_equal(res, expected)


def test_isclose_error():
    x = [1, 2, 3] * cm
    y = [1, 2, 3] * g
    with pytest.raises(UnitConversionError):
        np.isclose(x, y)


def test_linspace():
    res = np.linspace(1 * cm, 11 * cm, 10)
    assert type(res) is unyt_array
    assert res.units == cm


def test_linspace_with_retstep():
    res, step = np.linspace(1 * cm, 11 * cm, 10, retstep=True)
    assert type(res) is unyt_array
    assert res.units == cm
    assert type(step) is unyt_quantity
    assert step.units == cm


def test_logspace_with_units_raises():
    with pytest.raises(
        TypeError,
        match="The first two arguments to numpy.logspace must be dimensionless",
    ):
        np.logspace(1 * cm, 2 * cm)


def test_logspace_with_base():
    res = np.logspace(1, 2, base=10 * cm)
    assert type(res) is unyt_array
    assert res.units == cm


def test_logspace_with_dimless_input():
    res = np.logspace(1 * dimensionless, 2 * dimensionless)
    assert type(res) is unyt_array
    assert res.units == dimensionless


def test_geomspace():
    res = np.geomspace(1 * cm, 11 * cm, 10)
    assert type(res) is unyt_array
    assert res.units == cm


def test_full_like():
    x = [1, 2, 3] * cm
    res = np.full_like(x, 6 * cm)
    assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.parametrize(
    "func",
    [
        np.empty_like,
        np.zeros_like,
        np.ones_like,
    ],
)
def test_x_like(func):
    x = unyt_array([1, 2, 3], cm, dtype="float32")
    res = func(x)
    assert type(res) is unyt_array
    assert res.units == x.units
    assert res.shape == x.shape
    assert res.dtype == x.dtype


def test_copy():
    x = [1, 2, 3] * cm
    y = np.copy(x)
    # by default, subok=False, so we shouldn't
    # expect a unyt_array without switching this arg
    assert type(y) is np.ndarray


def test_copy_subok():
    x = [1, 2, 3] * cm
    y = np.copy(x, subok=True)
    assert type(y) is unyt_array
    assert y.units == cm


def test_copyto():
    x = [1, 2, 3] * cm
    y = np.empty_like(x)
    np.copyto(y, x)
    assert type(y) is unyt_array
    assert y.units == cm
    np.testing.assert_array_equal(x, y)


def test_copyto_edge_cases():
    x = [1, 2, 3] * cm
    y = [1, 2, 3] * g
    # copying to an array with a different unit is supported
    # to be in line with how we treat the 'out' param in most
    # numpy operations
    np.copyto(y, x)
    assert type(y) is unyt_array
    assert y.units == cm

    y = np.empty_like(x.view(np.ndarray))
    np.copyto(y, x)
    assert type(y) is np.ndarray


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="astype is new in numpy 2.0"
)
def test_astype():
    x = np.array([1, 2, 3], dtype="int64") * cm
    res = np.astype(x, "int32")
    assert type(res) is unyt_array
    assert res.units == cm


def test_meshgrid():
    x = [1, 2, 3] * cm
    y = [1, 2, 3] * s
    x2d, y2d = np.meshgrid(x, y)
    assert type(x2d) is unyt_array
    assert type(y2d) is unyt_array
    assert x2d.units == cm
    assert y2d.units == s


@pytest.mark.parametrize(
    "func, args, kwargs",
    [
        (np.transpose, (), {}),
        (np.reshape, ((9, 2),), {}),
        (np.resize, ((3, 6),), {}),
        (np.expand_dims, (0,), {}),
        (np.squeeze, (), {}),
        (np.swapaxes, (0, 1), {}),
        (np.moveaxis, (0, 2), {}),
        (np.rot90, (), {}),
        (np.roll, (3,), {}),
        (np.rollaxis, (2,), {}),
        (np.flip, (), {}),
        (np.fliplr, (), {}),
        (np.flipud, (), {}),
        (np.broadcast_to, ((1, 1, 2, 3, 3),), {"subok": True}),
        (np.delete, (0, 1), {}),
        (np.partition, (2,), {}),
    ],
)
def test_reshaper(func, args, kwargs):
    x = [
        [
            [
                [1, 2, 3],
                [4, 5, 6],
                [7, 8, 9],
            ],
            [
                [10, 11, 12],
                [13, 14, 15],
                [16, 17, 18],
            ],
        ]
    ] * cm
    y = func(x, *args, **kwargs)
    assert type(y) is unyt_array
    assert y.units == cm


def test_broadcast_arrays():
    x = [1, 2, 3] * cm
    y = [
        4,
    ] * g
    res = np.broadcast_arrays(x, y, subok=True)
    assert all(type(_) is unyt_array for _ in res)


@pytest.mark.parametrize(
    "func_name, args",
    [
        ("split", (3, 2)),
        ("dsplit", (3,)),
        ("hsplit", (2,)),
        ("vsplit", (1,)),
        ("array_split", (3,)),
        pytest.param(
            "unstack",
            (),
            marks=pytest.mark.skipif(
                NUMPY_VERSION < Version("2.1.0dev0"),
                reason="np.unstack is new in NumPy 2.1",
            ),
        ),
    ],
)
def test_xsplit(func_name, args):
    func = getattr(np, func_name)
    x = [
        [
            [
                [1, 2, 3],
                [4, 5, 6],
                [7, 8, 9],
            ],
            [
                [10, 11, 12],
                [13, 14, 15],
                [16, 17, 18],
            ],
        ]
    ] * cm
    y = func(x, *args)
    assert all(type(_) is unyt_array for _ in y)
    assert all(_.units == cm for _ in y)


@pytest.mark.parametrize(
    "func, expected_units",
    [
        (np.prod, cm**9),
        (np.var, cm**2),
        (np.std, cm),
        (np.nanprod, cm**9),
        (np.nansum, cm),
        (np.nanvar, cm**2),
        (np.nanstd, cm),
        (np.trace, cm),
    ],
)
def test_scalar_reducer(func, expected_units):
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = func(x)
    assert type(y) is unyt_quantity
    assert y.units == expected_units


@pytest.mark.parametrize("func", [np.prod, np.nanprod])
def test_prod_with_axis(func):
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = func(x, axis=0)
    assert type(y) is unyt_array
    assert y.units == cm**3


@pytest.mark.parametrize(
    "func",
    [
        np.percentile,
        np.quantile,
        np.nanpercentile,
        np.nanquantile,
    ],
)
def test_percentile(func):
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = func(x, 1)
    assert type(y) is unyt_quantity
    assert y.units == cm


@pytest.mark.parametrize(
    "func",
    [
        np.diag,
        np.diagflat,
        np.diagonal,
    ],
)
def test_diagx(func):
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = func(x)
    assert type(y) is unyt_array
    assert y.units == cm


def test_fix():
    y = np.fix(1.2 * cm)
    assert y == 1.0 * cm


def test_linalg_matrix_power():
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = np.linalg.matrix_power(x, 2)
    assert type(y) is unyt_array
    assert y.units == cm**2


def test_linalg_det():
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    y = np.linalg.det(x)
    assert type(y) is unyt_quantity
    assert y.units == cm ** (len(x))


def test_linalg_cholesky():
    x = np.eye(3) * cm
    y = np.linalg.cholesky(x)
    assert type(y) is unyt_array
    assert y.units == cm


def test_linalg_lstsq():
    a = np.eye(3) * cm
    b = np.ones(3).T * g
    # setting rcond explicitly to avoid a FutureWarning
    # see https://numpy.org/doc/stable/reference/generated/numpy.linalg.lstsq.html
    x, residuals, rank, s = np.linalg.lstsq(a, b, rcond=-1)

    assert type(x) is unyt_array
    assert x.units == g / cm
    assert type(residuals) is unyt_array
    assert residuals.units == g / cm
    assert type(s) is unyt_array
    assert s.units == cm


def test_linalg_multi_dot():
    a = np.eye(3) * cm
    b = np.eye(3) * g
    c = np.eye(3) * s
    res = np.linalg.multi_dot([a, b, c])
    assert type(res) is unyt_array
    assert res.units == cm * g * s


def test_linalg_qr():
    x = [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ] * cm
    q, r = np.linalg.qr(x)
    assert type(q) is unyt_array
    assert q.units == cm
    assert type(r) is unyt_array
    assert r.units == cm


@pytest.mark.parametrize("func", [np.linalg.solve, np.linalg.tensorsolve])
def test_linalg_solve(func):
    a = np.eye(3) * cm
    b = np.ones(3).T * g

    x = func(a, b)
    assert type(x) is unyt_array
    assert x.units == g / cm


def is_any_dimless(x) -> bool:
    return (not hasattr(x, "units")) or x.units.is_dimensionles


def test_linalg_cond():
    a = np.eye(3) * cm
    res = np.linalg.cond(a)
    assert is_any_dimless(res)


@pytest.mark.parametrize("func", [np.linalg.eig, np.linalg.eigh])
def test_eig(func):
    a = np.eye(3) * cm
    w, v = func(a)
    assert type(w) is unyt_array
    assert w.units == cm
    assert is_any_dimless(v)


@pytest.mark.parametrize("func", [np.linalg.eigvals, np.linalg.eigvalsh])
def test_eigvals(func):
    a = np.eye(3) * cm
    w = func(a)
    assert type(w) is unyt_array
    assert w.units == cm


def test_savetxt(tmp_path):
    a = [1, 2, 3] * cm
    with pytest.warns(
        UserWarning,
        match=re.escape(
            "numpy.savetxt does not preserve units, "
            "and will only save the raw numerical data from the unyt_array object.\n"
            "If this is the intended behaviour, call `numpy.savetxt(file, arr.d)` "
            "to silence this warning.\n"
            "If you want to preserve units, use `unyt.savetxt` "
            "(and `unyt.loadtxt`) instead."
        ),
    ):
        np.savetxt(tmp_path / "savefile.npy", a)

    # check that doing what the warning says doesn't trigger any other warning
    np.savetxt(tmp_path / "savefile.npy", a.d)


def test_apply_along_axis():
    a = np.eye(3) * cm
    ret = np.apply_along_axis(lambda x: x * cm, 0, a)
    assert type(ret) is unyt_array
    assert ret.units == cm**2


@pytest.mark.parametrize("axes, expected_units", [((0, 1), cm**4), ((0,), cm**2)])
def test_apply_over_axes(axes, expected_units):
    # the user-supplied function must be trusted to treat units
    # sensibly (mainly that it doesn't give a mix of units across
    # the resulting array), but we can check that units are
    # propagated correctly for well-behaved functions.
    a = np.eye(3) * cm
    ret = np.apply_over_axes(lambda x, axis: x[axis] ** 2, a, axes)
    assert isinstance(ret, unyt_array)  # could be unyt_quantity
    assert ret.units == expected_units


def test_array_equal():
    a = [1, 2, 3] * cm
    b = [1, 2, 3] * cm
    c = [1, 2, 3] * km
    assert np.array_equal(a, b)
    assert not np.array_equal(a, c)


def test_array_equiv():
    a = [1, 2, 3] * cm
    b = [1, 2, 3] * cm
    c = [1, 2, 3] * km
    d = [[1, 2, 3]] * cm
    assert np.array_equiv(a, b)
    assert np.array_equiv(a, d)
    assert not np.array_equiv(a, c)


def test_common_type():
    a = np.array([1, 2, 3], dtype="float32") * cm
    b = np.array([1, 2, 3], dtype="float64") * cm
    dtype = np.common_type(a, b)
    assert dtype == np.dtype("float64")


def test_result_type():
    scalar = 3 * cm
    array = np.arange(7, dtype="i1")
    if NUMPY_VERSION >= Version("2.0.0dev0"):
        # promotion rules vary under NEP 50. The default behaviour is different
        # in numpy 2.0 VS numpy 1.x
        # see https://github.com/numpy/numpy/pull/23912
        # see https://numpy.org/neps/nep-0050-scalar-promotion.html
        expected_dtype = scalar.dtype
    else:
        expected_dtype = array.dtype

    assert np.result_type(scalar, array) == expected_dtype


@pytest.mark.parametrize(
    "func",
    [
        np.diff,
        np.ediff1d,
        np.gradient,
        np.ptp,
    ],
)
@pytest.mark.parametrize("input_units, output_units", [(cm, cm), (K, delta_degC)])
def test_deltas(func, input_units, output_units):
    x = np.arange(0, 4) * input_units
    res = func(x)
    assert isinstance(res, unyt_array)
    assert res.units == output_units


@pytest.mark.parametrize(
    "func_name",
    [
        "cumsum",
        "nancumsum",
        pytest.param(
            "cumulative_sum",
            marks=pytest.mark.skipif(
                NUMPY_VERSION < Version("2.1.0dev0"),
                reason="np.cumulative_sum is new in NumPy 2.1",
            ),
        ),
    ],
)
def test_cumsum(func_name):
    a = [1, 2, 3] * cm
    func = getattr(np, func_name)
    res = func(a)
    assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.parametrize(
    "func_name",
    [
        "cumprod",
        "nancumprod",
        pytest.param(
            "cumulative_prod",
            marks=pytest.mark.skipif(
                NUMPY_VERSION < Version("2.1.0dev0"),
                reason="np.cumulative_prod is new in NumPy 2.1",
            ),
        ),
    ],
)
def test_cumprod(func_name):
    a = [1, 2, 3] * cm
    func = getattr(np, func_name)
    with pytest.raises(
        UnytError,
        match=re.escape(
            r"numpy.cumprod (and other cumulative product function) cannot be used "
            r"with a unyt_array as all return elements should (but cannot) "
            r"have different units."
        ),
    ):
        func(a)


def test_bincount():
    a = [1, 2, 3] * cm
    res = np.bincount(a)
    assert type(res) is np.ndarray


def test_unique():
    a = [1, 2, 3] * cm
    res = np.unique(a)
    assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="unique_all is new in numpy 2.0"
)
def test_unique_all():
    q = np.arange(9).reshape(3, 3) * cm
    values, indices, inverse_indices, counts = np.unique(
        q,
        return_index=True,
        return_inverse=True,
        return_counts=True,
        equal_nan=False,
    )
    res = np.unique_all(q)
    assert len(res) == 4
    assert_array_equal_units(res.values, values)
    assert_array_equal_units(res.indices, indices)
    assert_array_equal_units(res.inverse_indices, inverse_indices)
    assert_array_equal_units(res.counts, counts)


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="unique_counts is new in numpy 2.0"
)
def test_unique_counts():
    q = np.arange(9).reshape(3, 3) * cm
    values, counts = np.unique(
        q,
        return_counts=True,
        equal_nan=False,
    )
    res = np.unique_counts(q)
    assert len(res) == 2
    assert_array_equal_units(res.values, values)
    assert_array_equal_units(res.counts, counts)


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="unique_inverse is new in numpy 2.0"
)
def test_unique_inverse():
    q = np.arange(9).reshape(3, 3) * cm
    values, inverse_indices = np.unique(
        q,
        return_inverse=True,
        equal_nan=False,
    )
    res = np.unique_inverse(q)
    assert len(res) == 2
    assert_array_equal_units(res.values, values)
    assert_array_equal_units(res.inverse_indices, inverse_indices)


@pytest.mark.skipif(
    NUMPY_VERSION < Version("2.0.0dev0"), reason="unique_values is new in numpy 2.0"
)
def test_unique_values():
    q = np.arange(9).reshape(3, 3) * cm
    values = np.unique(q, equal_nan=False)
    res = np.unique_values(q)

    # np.unique_values' output is not guaranteed to be sorted,
    # so exact results may differ.
    # see https://github.com/numpy/numpy/issues/28493#issuecomment-2721303048
    res.sort()
    assert_array_equal_units(res, values)


@pytest.mark.parametrize("indices", [[0, 1], 0])
def test_take(indices):
    a = [1, 2, 3] * cm
    res = np.take(a, indices)
    if res.ndim == 0:
        assert type(res) is unyt_quantity
    else:
        assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.parametrize("indices", [[0, 1], 0])
def test_ndarray_take(indices):
    a = [1, 2, 3] * cm
    res = a.take(indices)
    if res.ndim == 0:
        assert type(res) is unyt_quantity
    else:
        assert type(res) is unyt_array
    assert res.units == cm


def test_pad():
    a = [1, 2, 3] * cm
    res = np.pad(a, [0, 1])
    assert type(res) is unyt_array
    assert res.units == cm


def test_sinc():
    a = [1, 2, 3] * cm
    res = np.sinc(a)
    # we *want* this one to ignore units
    assert type(res) is np.ndarray


def test_choose_mixed_units():
    choices = [[1, 2, 3] * cm, [4, 5, 6] * km]
    with pytest.raises(UnitInconsistencyError):
        np.choose([1, 0, 1], choices=choices)


def test_choose():
    choices = [[1, 2, 3] * cm, [4, 5, 6] * cm]
    res = np.choose([1, 0, 1], choices=choices)
    assert type(res) is unyt_array
    assert res.units == cm


def test_extract():
    a = [1, 2, 3] * cm
    res = np.extract(a > 1 * cm, a)
    assert type(res) is unyt_array
    assert res.units == cm


def test_fill_diagonal_mixed_units():
    a = np.zeros(9).reshape((3, 3)) * cm
    with pytest.raises(UnitInconsistencyError):
        np.fill_diagonal(a, 1 * km)


@pytest.mark.parametrize("val", [1 * cm, 1])
def test_fill_diagonal(val):
    a = np.zeros(9).reshape((3, 3)) * cm
    np.fill_diagonal(a, val)
    assert type(a) is unyt_array
    assert a.units == cm


def test_insert_mixed_units():
    a = [1, 2, 3] * cm
    with pytest.raises(UnitInconsistencyError):
        np.insert(a, 1, 42 * km)


@pytest.mark.parametrize("val", [42, 42 * cm])
def test_insert(val):
    a = [1, 2, 3] * cm
    res = np.insert(a, 1, val)
    assert type(res) is unyt_array
    assert res.units == cm


def test_isin_mixed_units():
    a = [1, 2, 3] * cm
    with pytest.raises(UnitInconsistencyError):
        np.isin(1, a)


def test_isin():
    a = [1, 2, 3] * cm
    assert np.isin(1 * cm, a)


@pytest.mark.skipif(
    NUMPY_VERSION >= Version("2.4.0.dev0"), reason="in1d is removed in numpy 2.4"
)
@pytest.mark.filterwarnings("ignore:`in1d` is deprecated. Use `np.isin` instead.")
def test_in1d_mixed_units():
    a = [1, 2, 3] * cm
    with pytest.raises(UnitInconsistencyError):
        np.in1d([1, 2], a)  # noqa: NPY201


@pytest.mark.skipif(
    NUMPY_VERSION >= Version("2.4.0.dev0"), reason="in1d is removed in numpy 2.4"
)
@pytest.mark.filterwarnings("ignore:`in1d` is deprecated. Use `np.isin` instead.")
def test_in1d():
    a = [1, 2, 3] * cm
    b = [1, 2] * cm
    assert np.all(np.in1d(b, a))  # noqa: NPY201


def test_place_mixed_units():
    arr = np.arange(6).reshape(2, 3) * cm
    with pytest.raises(UnitInconsistencyError):
        np.place(arr, arr > 2, [44, 55])


def test_place():
    arr = np.arange(6).reshape(2, 3) * cm
    np.place(arr, arr > 2, [44, 55] * cm)
    assert type(arr) is unyt_array
    assert arr.units == cm


def test_put_mixed_units():
    arr = np.arange(6).reshape(2, 3) * cm
    with pytest.raises(UnitInconsistencyError):
        np.put(arr, [1, 2], [44, 55])


def test_put():
    arr = np.arange(6).reshape(2, 3) * cm
    np.put(arr, [1, 2], [44, 55] * cm)
    assert type(arr) is unyt_array
    assert arr.units == cm


def test_put_along_axis_mixed_units():
    arr = np.arange(6).reshape(2, 3) * cm
    with pytest.raises(UnitInconsistencyError):
        np.put_along_axis(arr, np.array([[1, 2], [0, 1]]), [44, 55], 1)


def test_put_along_axis():
    arr = np.arange(6).reshape(2, 3) * cm
    np.put_along_axis(arr, np.array([[1, 2], [0, 1]]), [44, 55] * cm, 1)
    assert type(arr) is unyt_array
    assert arr.units == cm


def test_putmask_mixed_units():
    arr = np.arange(6, dtype=np.int_).reshape(2, 3) * cm
    with pytest.raises(UnitInconsistencyError):
        np.putmask(arr, arr > 2 * cm, np.zeros_like(arr.d))


def test_putmask():
    arr = np.arange(6, dtype=np.int_).reshape(2, 3) * cm
    np.putmask(arr, arr > 2 * cm, np.zeros_like(arr))

    assert type(arr) is unyt_array
    assert arr.units == cm


def test_searchsorted_mixed_units():
    with pytest.raises(UnitInconsistencyError):
        np.searchsorted([1, 2, 3, 4, 5] * cm, 3 * km)


@pytest.mark.parametrize("val", [3 * cm, 3])
def test_searchsorted(val):
    res = np.searchsorted([1, 2, 3, 4, 5] * cm, val)
    assert res == 2


@pytest.mark.parametrize(
    "choicelist_gen, default",
    [
        ([lambda x: x**3, lambda x: x**2], 34 * cm),  # invalid choicelist
        ([lambda x: x, lambda x: x + 3 * x.units], 34),  # invalid default
    ],
)
def test_select_mixed_units(choicelist_gen, default):
    a = [1, 2, 3, 4, 5, 6] * cm
    with pytest.raises(UnitInconsistencyError):
        np.select(
            [a > 3 * cm, a < 3 * cm], [f(a) for f in choicelist_gen], default=default
        )


def test_select():
    a = [1, 2, 3, 4, 5, 6] * cm
    res = np.select([a > 3 * cm, a < 3 * cm], [a, a + 3 * cm], default=34 * cm)
    assert_array_equal_units(res, [4, 5, 34, 4, 5, 6] * cm)


def test_setdiff1d_mixed_units():
    a = [1, 2, 3] * cm
    b = [0, 1, 2]
    with pytest.raises(UnitInconsistencyError):
        np.setdiff1d(a, b)


def test_setdiff1d():
    a = [1, 2, 3] * cm
    b = [0, 1, 2] * cm
    res = np.setdiff1d(a, b)
    assert_array_equal_units(res, [3] * cm)


def test_setxor1d_mixed_units():
    a = [1, 2, 3] * cm
    b = [0, 1, 2]
    with pytest.raises(UnitInconsistencyError):
        np.setxor1d(a, b)


def test_setxor1d():
    a = [1, 2, 3] * cm
    b = [0, 1, 2] * cm
    res = np.setxor1d(a, b)
    assert_array_equal_units(res, [0, 3] * cm)


def test_clip_mixed_units():
    a = [1, 2, 3, 4, 5, 6] * cm
    with pytest.raises(UnitInconsistencyError):
        np.clip(a, 3 * cm, 4)


@pytest.mark.parametrize("vmin,vmax", [(3 * cm, 4 * cm), (3, 4)])
def test_clip(vmin, vmax):
    a = [1, 2, 3, 4, 5, 6] * cm
    res = np.clip(a, vmin, vmax)
    assert_array_equal_units(res, [3, 3, 3, 4, 4, 4] * cm)


def test_where_mixed_units():
    x = [-1, 2, -3] * cm
    y = [0, 0, 0]
    with pytest.raises(UnitInconsistencyError):
        np.where(x > y, x, y)


def test_where_single_arg():
    x = [0, 2, -1, 0, 1] * cm
    res = np.where(x)
    assert isinstance(res, tuple)
    assert len(res) == 1
    assert type(res[0]) is np.ndarray
    np.testing.assert_array_equal(res[0], [1, 2, 4])


def test_where_xy():
    x = [-1, 2, -3] * cm
    y = [0, 0, 0] * cm
    res = np.where(x > y, x, y)
    assert type(res) is unyt_array
    assert res.units == cm


@pytest.mark.parametrize(
    "func",
    [
        np.imag,
        np.real,
        np.real_if_close,
    ],
)
def test_complex_reductions(func):
    a = [1 + 1j for _ in range(3)] * A
    res = func(a)
    assert type(res) is unyt_array
    assert res.units == A


@pytest.mark.parametrize(
    "func",
    [np.tril, np.triu],
)
def test_triangles(func):
    a = np.eye(4) * cm
    res = func(a)
    assert type(res) is unyt_array
    assert res.units == cm


def test_einsum():
    a = np.eye(4) * cm

    # extract diagonal
    res = np.einsum("ii->i", a)
    assert type(res) is unyt_array
    assert res.units == cm

    # sum diagonal elements, the result should be a scalar
    res = np.einsum("ii", a)
    assert type(res) is unyt_quantity
    assert res.units == cm


def test_convolve():
    a = [1, 2, 3] * cm
    v = [4, 5, 6] * s
    res = np.convolve(a, v)
    assert type(res) is unyt_array
    assert res.units == cm * s


def test_correlate():
    a = [1, 2, 3] * cm
    v = [4, 5, 6] * s
    res = np.correlate(a, v)
    assert type(res) is unyt_array
    assert res.units == cm * s


def test_tensordot():
    a = np.arange(60.0).reshape(3, 4, 5) * cm
    b = np.arange(24.0).reshape(4, 3, 2) * s
    res = np.tensordot(a, b, axes=([1, 0], [0, 1]))
    assert type(res) is unyt_array
    assert res.units == cm * s


def test_compress():
    a = [1, 2, 3] * cm
    res = np.compress(a > 1, a)
    assert type(res) is unyt_array
    assert res.units == cm

    np.compress(a > 1, a, out=res)
    assert type(res) is unyt_array
    assert res.units == cm

    np.compress(a > 1, a, out=res.view(np.ndarray))
    assert type(res) is unyt_array
    assert res.units == cm


def test_take_along_axis():
    a = np.array([[10, 30, 20], [60, 40, 50]]) * cm
    ai = np.argsort(a, axis=1)
    res = np.take_along_axis(a, ai, axis=1)
    assert type(res) is unyt_array
    assert res.units == cm


def test_unwrap():
    phase = np.linspace(0, np.pi, num=5) * rad
    phase[3:] += np.pi * rad
    res = np.unwrap(phase)
    assert type(res) is unyt_array
    assert res.units == rad


def test_interp():
    _x = np.array([1.1, 2.2, 3.3])
    _xp = np.array([1, 2, 3])
    _fp = np.array([4, 8, 12])

    # any of the three input array-like might be unitful
    # let's test all relevant combinations
    # return type should match fp's

    with pytest.raises(UnitInconsistencyError):
        np.interp(_x * cm, _xp, _fp)

    with pytest.raises(UnitInconsistencyError):
        res = np.interp(_x, _xp * cm, _fp)

    res = np.interp(_x * cm, _xp * cm, _fp)
    assert type(res) is np.ndarray

    res = np.interp(_x, _xp, _fp * K)
    assert type(res) is unyt_array
    assert res.units == K

    res = np.interp(_x * cm, _xp * cm, _fp * K)
    assert type(res) is unyt_array
    assert res.units == K


@pytest.mark.parametrize(
    "target, helper",
    sorted(
        HANDLED_FUNCTIONS.items(),
        key=lambda items: items[0].__name__,
    ),
    ids=lambda func: func.__name__,
)
class TestFunctionHelpersSignatureCompatibility:
    """
    Check that a helper function's signature is *at least* as flexible
    as the helped (target) function's. E.g., any argument that is allowed positionally,
    or as keyword, by the target must be re-exposed *somehow* by the helper.
    We explicitly allow helper's signature to be *more* flexible than the target signature
    by allowing *args and **kwargs catch-all arguments, which we use to limit code
    duplication, and also help with forward and backward compatibility.
    See https://github.com/astropy/astropy/issues/15703
    """

    # this test class is adapted from astropy.units

    @staticmethod
    def have_catchall_argument(parameters, kind) -> bool:
        return any(p.kind is kind for p in parameters.values())

    @staticmethod
    def get_param_group(parameters, kinds: list) -> list[str]:
        return [name for name, p in parameters.items() if p.kind in kinds]

    def test_all_arguments_reexposed(self, target, helper):
        try:
            sig_target = inspect.signature(target)
        except ValueError:
            pytest.skip("Non Python function cannot be inspected at runtime")

        params_target = sig_target.parameters
        sig_helper = inspect.signature(helper)
        params_helper = sig_helper.parameters

        have_args_helper = self.have_catchall_argument(params_helper, VAR_POSITIONAL)
        have_kwargs_helper = self.have_catchall_argument(params_helper, VAR_KEYWORD)

        args_helper = list(params_helper.items())

        pos_helper = 0
        for nt, pt in params_target.items():
            kt = pt.kind
            if kt in (POSITIONAL_ONLY, POSITIONAL_OR_KEYWORD):
                assert pos_helper < len(args_helper), (
                    "helper's signature is too short; "
                    "some arguments are not properly re-exposed"
                )
                nh, ph = args_helper[pos_helper]
                if (kh := ph.kind) is not VAR_POSITIONAL:
                    assert nh == nt, f"argument {nt!r} isn't re-exposed as positional"
                    assert kh is kt, (
                        f"helper is not re-exposing argument {nt!r} properly: "
                        f"expected {kt}, got {kh}"
                    )
                    pos_helper += 1
                    continue

            if kt in (KEYWORD_ONLY, POSITIONAL_OR_KEYWORD):
                if nt in params_helper:
                    kh = params_helper[nt].kind
                    assert kh is kt, (
                        f"helper is not re-exposing argument {nt!r} properly: "
                        f"expected {kt}, got {kh}"
                    )
                elif kt is KEYWORD_ONLY:
                    assert have_kwargs_helper, (
                        f"argument {nt!r} is not re-exposed as keyword"
                    )
                elif kt is POSITIONAL_OR_KEYWORD:
                    assert have_args_helper and have_kwargs_helper, (
                        f"argument {nt!r} is not re-exposed as positional-or-keyword"
                    )
            elif kt is VAR_POSITIONAL:
                assert have_args_helper, "helper is missing a catch-all *args argument"
            elif kt is VAR_KEYWORD:
                assert have_kwargs_helper, (
                    "helper is missing a catch-all **kwargs argument"
                )

    def test_known_arguments(self, target, helper):
        # validate that all exposed arguments map to something in the target
        try:
            sig_target = inspect.signature(target)
        except ValueError:
            pytest.skip("Non Python function cannot be inspected at runtime")

        params_target = sig_target.parameters
        sig_helper = inspect.signature(helper)
        params_helper = sig_helper.parameters

        for kind in (POSITIONAL_ONLY, POSITIONAL_OR_KEYWORD):
            args_target = self.get_param_group(params_helper, [kind])
            args_helper = self.get_param_group(params_helper, [kind])

            if (nhelper := len(args_helper)) > (ntarget := len(args_target)):
                unknown: list[str] = args_helper[ntarget:]
                raise AssertionError(
                    f"Found unknown {kind} parameter(s) "
                    "in helper's signature: "
                    f"{unknown}, at position(s) {list(range(ntarget, nhelper))}"
                )

        # keyword-allowed
        keyword_allowed_target = set(
            self.get_param_group(params_target, [KEYWORD_ONLY, POSITIONAL_OR_KEYWORD])
        )
        keyword_allowed_helper = set(
            self.get_param_group(params_helper, [KEYWORD_ONLY, POSITIONAL_OR_KEYWORD])
        )

        # additional private keyword-only argument are allowed because
        # they are only intended for testing purposes.
        # For instance, quantile has such a parameter '_q_unit'
        keyword_allowed_helper = {
            name for name in keyword_allowed_helper if not name.startswith("_")
        }

        diff = keyword_allowed_helper - keyword_allowed_target
        assert not diff, (
            "Found some keyword-allowed parameters in helper "
            f"that are unknown to target: {diff}"
        )

        # finally, check that default values are correctly replicated
        for name, ph in params_helper.items():
            if name not in params_target:
                # In a few cases, the helper defines names that are not in
                # the target (e.g., a private name like _q_unit in quantile,
                # or a *args, **kwargs that captures further arguments
                # that do not matter. We let such cases slip by.
                continue
            pt = params_target[name]
            assert ph.default == pt.default, (
                f"Default value mismatch for argument {name!r}. "
                f"Helper has {ph.default!r}, target has {pt.default!r}"
            )



import pickle
from collections import defaultdict

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from unyt import unyt_array, unyt_quantity
from unyt._on_demand_imports import _dask as dask

if not dask.__is_available__:
    pytest.skip("dask isn't installed", allow_module_level=True)

from unyt.dask_array import (
    _create_with_quantity,
    _use_unary_decorator,
    reduce_with_units,
    unyt_dask_array,
    unyt_from_dask,
)
from unyt.exceptions import UnitOperationError
from unyt.unit_symbols import cm, g, m


def test_unyt_dask_creation():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    assert type(x_da) is unyt_dask_array
    assert x_da.units == m
    assert type(x_da.compute()) is unyt_array

    x_da = _create_with_quantity(x, unyt_quantity(1, m))
    assert type(x_da) is unyt_dask_array
    assert x_da.units == m
    assert type(x_da.compute()) is unyt_array

    x_dask = x_da.to_dask()
    assert_array_equal(x.compute(), x_dask.compute())


def test_unyt_dask_slice():
    # tests __getitem__
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    slc = x_da[:, 0]
    assert slc.units == m
    assert slc.compute().units == m
    assert type(slc.compute()) is unyt_array


def test_unyt_set():
    # tests __setitem__
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_da[:, 0] = 3.0
    assert (x_da[:, 0].compute() == 3.0).sum() == 10


def test_unit_conversions():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_da = x_da.to(cm)
    assert type(x_da) is unyt_dask_array
    assert x_da.units == cm
    assert x_da.compute().units == cm
    assert x_da[0, 0].compute().value == 100

    x_da_2 = unyt_from_dask(x, m)
    result = x_da + x_da_2
    assert type(result) is unyt_dask_array
    assert result.units == m
    assert result.compute().units == m

    x_da_2 = unyt_from_dask(x, "g")
    with pytest.raises(UnitOperationError):
        x_da + x_da_2


def test_conversion_to_dask():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_again = x_da.to_dask()
    assert_array_equal(x.compute(), x_again.compute())
    assert type(x_again) is type(x)

    result = np.isfinite(x_da)  # should return plain dask array
    assert type(result) is dask.array.core.Array


def unary_test(the_func, unyt_dask_obj, unyt_array_in, *args, **kwargs):
    # the_func can be a numpy ufunc or dask.array numpy ufunc implementation

    result_delay = the_func(unyt_dask_obj, *args, **kwargs)
    correct_unyt = the_func(unyt_array_in, *args, **kwargs)

    assert result_delay.units == correct_unyt.units  # units should already match
    unary_result_test(result_delay, correct_unyt)


def unary_result_test(dask_unyt_delayed, correct_unyt):
    # computes a delayed dask_unyt_array and compares resulting units and values
    result = dask_unyt_delayed.compute()
    assert result.units == correct_unyt.units
    assert type(result) is type(correct_unyt)
    # value comparison:
    if type(result) is unyt_array:
        assert_array_equal(result, correct_unyt)
    else:
        assert result == correct_unyt


def test_unary():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_unyt = unyt_array(np.ones((10, 10)), m)

    for unary_op in [np.sqrt, dask.array.sqrt]:
        unary_test(unary_op, x_da, x_unyt)

    # some of the daskified ufunc attributes need to be called directly
    unary_result_test(x_da.sum(), x_unyt.sum())
    unary_result_test(x_da.min(), x_unyt.min())
    unary_result_test(x_da.max(), x_unyt.max())
    unary_result_test(x_da.mean(), x_unyt.mean())
    unary_result_test(x_da.std(), x_unyt.std())
    unary_result_test(x_da.cumsum(0), x_unyt.cumsum(0))
    unary_result_test(abs(x_da), abs(x_unyt))  # __abs__


@pytest.mark.parametrize(
    "logical_op", ["__lt__", "__le__", "__gt__", "__ge__", "__eq__", "__ne__"]
)
@pytest.mark.parametrize("other", [None, unyt_quantity(2, m)])
def test_logical(logical_op, other):
    def check_operator(arg1, arg2):
        # comparisons should return plain dask arrays
        func = getattr(arg1, logical_op)
        result = func(arg2)
        assert type(result) is dask.array.Array

    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    if other is None:
        other = 2 * x_da  # test another unyt-dask array
    check_operator(x_da, other)
    check_operator(other, x_da)


def test_binary():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x2 = dask.array.full((10, 10), 2, chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_da_2 = unyt_from_dask(x2, g)

    # multiplications
    result = x_da * x_da_2
    assert result.units == m * g
    result = x_da_2 * x_da
    assert result.units == m * g
    result = x_da_2 * 2
    assert result.units == g
    result = 2 * x_da_2
    assert result.units == g
    result = x_da_2 * unyt_quantity(2, "m")
    assert result.units == m * g
    result = unyt_quantity(2, "m") * x_da_2
    assert result.units == m * g

    # divisions
    result = x_da_2 / x_da
    assert result.units == g / m
    result = x_da / x_da_2
    assert result.units == m / g
    result = x_da_2 / 2
    assert result.units == g
    result = 2 / x_da_2  # __rtruediv__
    assert result.units == g**-1
    result = x_da_2 / unyt_quantity(2, "m")
    assert result.units == g / m
    result = unyt_quantity(2, "m") / x_da_2
    assert result.units == m / g

    result = x_da**2
    assert result.units == m * m


def test_addition():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x2 = dask.array.full((10, 10), 2, chunks=(2, 2))
    x_da = unyt_from_dask(x, m)

    # two unyt_dask_array objects, any order, any units with same dimension
    x_da_2 = unyt_from_dask(x2, m)
    result = x_da + x_da_2
    assert result.units == m
    result = x_da_2 + x_da
    assert result.units == m
    x_da_3 = unyt_from_dask(x2, "cm")
    result = x_da + x_da_3
    assert result.units == m
    result = x_da_3 + x_da
    assert result.units == m

    # one unyt_dask_array, one unyt_quantity, any order, any units with same dim
    result = x_da + unyt_quantity(1, "m")
    assert result.units == m
    result = unyt_quantity(1, "m") + x_da
    assert result.units == m
    result = unyt_quantity(100, "cm") + x_da  # test same dimensions
    assert result.units == m
    assert result.max().compute() == unyt_quantity(2, "m")
    result = x_da + unyt_quantity(100, "cm")  # test same dimensions
    assert result.units == m
    assert result.max().compute() == unyt_quantity(200, "cm")


def test_subtraction():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x2 = dask.array.full((10, 10), 2, chunks=(2, 2))
    x_da = unyt_from_dask(x, m)

    # two unyt_dask_array objects, any order, any units with same dimension
    x_da_2 = unyt_from_dask(x2, m)
    result = x_da - x_da_2
    assert result.units == m
    result = x_da_2 - x_da
    assert result.units == m
    x_da_3 = unyt_from_dask(x2, "cm")
    result = x_da - x_da_3
    assert result.units == m
    result = x_da_3 - x_da
    assert result.units == m

    # one unyt_dask_array, one unyt_quantity, any order, any units with same dim
    result = x_da - unyt_quantity(1, "m")
    assert result.units == m
    result = unyt_quantity(1, "m") - x_da
    assert result.units == m
    result = unyt_quantity(100, "cm") - x_da  # test same dimensions
    assert result.units == m
    assert result.max().compute() == unyt_quantity(0, "m")
    result = x_da - unyt_quantity(100, "cm")  # test same dimensions
    assert result.units == m
    assert result.max().compute() == unyt_quantity(0, "cm")


def test_unyt_type_result():
    # test that the return type of a compute is unyt_array or unyt_quantity when
    # appropriate

    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    x_unyt = unyt_array(np.ones((10, 10)), m)

    result = x_da.compute()
    assert type(result) is unyt_array
    assert_array_equal(result, x_unyt)

    result = x_da.min().compute()
    assert type(result) is unyt_quantity
    assert result == unyt_quantity(1, m)


_func_args = defaultdict(lambda: ())
_func_args["reshape"] = ((100, 1),)
_func_args["rechunk"] = ((5, 5),)
_func_args["cumsum"] = (0,)
_func_args["clip"] = (0, 2)
_func_args["swapaxes"] = (0, 1)
_func_args["repeat"] = (1, 0)
_func_args["astype"] = (int,)
_func_args["topk"] = (1,)


@pytest.mark.parametrize(
    ("da_func", "args"), [(f, _func_args[f]) for f in _use_unary_decorator]
)
def test_dask_passthroughs(da_func, args):
    # tests the array class functions that do not modify units
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    assert getattr(x_da, da_func)(*args).units == m


def test_repr():
    x = dask.array.ones((10, 10), chunks=(2, 2))
    x_da = unyt_from_dask(x, m)
    assert "unyt_dask_array" in x_da.__repr__()
    table_str = x_da._repr_html_()
    assert "Units" in table_str


@pytest.mark.parametrize(
    ("dask_func_str", "actual", "axis", "check_nan"),
    [
        ("mean", unyt_quantity(1.0, "m"), None, True),
        ("min", unyt_quantity(1.0, "m"), None, True),
        ("max", unyt_quantity(1.0, "m"), None, True),
        ("std", unyt_quantity(0.0, "m"), None, True),
        (
            "median",
            unyt_quantity(1.0, "m"),
            1,
            True,
        ),  # median requires an axis for dask
        ("var", unyt_quantity(0.0, "m**2"), None, True),
        ("sum", unyt_quantity(100.0, "m"), None, True),
        (
            "cumsum",
            unyt_array(np.ones((10, 10)), "m").cumsum(axis=1),
            1,
            True,
        ),
        ("average", unyt_quantity(1.0, "m"), None, False),
        ("diagonal", unyt_array(np.ones((10,)), "m"), None, False),
        ("unique", unyt_array([1.0], "m"), None, False),
    ],
)
def test_dask_array_reductions(dask_func_str, actual, axis, check_nan):
    extra_kwargs = {}
    if axis is not None:
        extra_kwargs["axis"] = axis

    func_strs = [
        dask_func_str,
    ]
    if check_nan:
        func_strs.append("nan" + dask_func_str)

    for func_str in func_strs:
        dask_func = getattr(dask.array, func_str)
        x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), m)

        result = reduce_with_units(dask_func, x_da, **extra_kwargs).compute()
        assert_array_equal(result, actual)


def test_bad_dask_array_reductions():
    x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), m)

    with pytest.raises(
        ValueError, match="could not deduce np equivalent of dask reduction"
    ):
        reduce_with_units(lambda: None, x_da).compute()


def test_prod():
    x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), m)
    assert x_da.prod().units == m**100


def test_pickle():
    x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), m)
    x_da2 = pickle.loads(pickle.dumps(x_da))
    assert_array_equal(x_da, x_da2)
    assert x_da.units == x_da2.units


def test_np_array_funcs():
    # test some numpy array functions to catch __array_function__

    x_da = unyt_from_dask(dask.array.ones((10, 10), chunks=(2, 2)), m)
    actual = np.sum(x_da).compute()
    assert actual == unyt_quantity(100, m)

    plain_dask = dask.array.arange(100, chunks=(20,))
    x_da = unyt_from_dask(plain_dask, m)
    actual = np.mean(x_da).compute()
    expected = unyt_quantity(np.mean(plain_dask).compute(), m)
    assert actual == expected

    actual = np.std(x_da[x_da > 10.0]).compute()
    expected = unyt_quantity(np.std(plain_dask[plain_dask > 10]).compute(), m)
    assert actual == expected

    actual = np.max(np.power(x_da, 2)).compute()
    expected = unyt_quantity(np.max(np.power(plain_dask, 2)).compute(), m * m)
    assert actual == expected



import os
import warnings
from copy import copy, deepcopy
from importlib.metadata import version

import numpy as np
import pytest
from packaging.version import Version

import unyt as u

from .sample_subclasses import ExtraAttributeError, subclass_uarray, subclass_uquantity

NUMPY_VERSION = Version(version("numpy"))
trapezoid_fname = "trapz" if NUMPY_VERSION < Version("2.0.0dev0") else "trapezoid"

savetxt_file = "saved_array.txt"


def getfunc(fname):
    """
    Helper for our tests: get the function handle from a name (possibly with attribute
    access).
    """
    func = np
    for attr in fname.split("."):
        func = getattr(func, attr)
    return func


def sub_arr(x, unit=u.Mpc):
    """
    Helper for our tests: turn an array into a subclass_uarray.
    """
    return subclass_uarray(x, unit, extra_attr=True)


def sub_q(x, unit=u.Mpc):
    """
    Helper for our tests: turn a scalar into a subclass_uquantity.
    """
    return subclass_uquantity(x, unit, extra_attr=True)


def arg_to_ua(arg):
    """
    Helper for our tests: recursively convert subclass_* in an argument (possibly an
    iterable) to their unyt_* equivalents.
    """
    if type(arg) in (list, tuple):
        return type(arg)([arg_to_ua(a) for a in arg])
    else:
        return to_ua(arg)


def to_ua(x):
    """
    Helper for our tests: turn a subclass_* object into its unyt_* equivalent.
    """
    return u.unyt_array(x) if hasattr(x, "extra_attr") else x


def check_result(x_s, x_u, ignore_values=False):
    """
    Helper for our tests: check that a result with subclass input matches what we
    expected based on the result with unyt input.

    We check:
     - that the type of the result makes sense, recursing if needed.
     - that the value of the result matches (unless ignore_values=False).
     - that the units match.
    """
    if x_u is None:
        assert x_s is None
        return
    if isinstance(x_u, str):
        assert isinstance(x_s, str)
        return
    if isinstance(x_u, type) or isinstance(x_u, np.dtype):
        assert x_u == x_s
        return
    if type(x_u) in (list, tuple):
        assert type(x_u) is type(x_s)
        assert len(x_u) == len(x_s)
        for x_c_i, x_u_i in zip(x_s, x_u):
            check_result(x_c_i, x_u_i)
            return
    # careful, unyt_quantity is a subclass of unyt_array:
    if isinstance(x_u, u.unyt_quantity):
        assert isinstance(x_s, subclass_uquantity)
    elif isinstance(x_u, u.unyt_array):
        assert isinstance(x_s, subclass_uarray) and not isinstance(
            x_s, subclass_uquantity
        )
    else:
        assert not isinstance(x_s, subclass_uarray)
        if not ignore_values:
            assert np.allclose(x_s, x_u)
        return
    assert x_s.units == x_u.units
    if not ignore_values:
        assert np.allclose(x_s.to_value(x_s.units), x_u.to_value(x_u.units))
    if isinstance(x_s, subclass_uarray):  # includes subclass_uquantity
        assert x_s.extra_attr is True
    return


class TestSubclassArrayInit:
    """
    Test different ways of initializing a subclass_uarray.
    """

    def test_init_from_ndarray(self):
        """
        Check initializing from a bare numpy array.
        """
        arr = subclass_uarray(np.ones(5), units=u.Mpc, extra_attr=True)
        assert hasattr(arr, "extra_attr")
        assert isinstance(arr, subclass_uarray)

    def test_init_from_list(self):
        """
        Check initializing from a list of values.
        """
        arr = subclass_uarray(
            [1, 1, 1, 1, 1],
            units=u.Mpc,
            extra_attr=True,
        )
        assert hasattr(arr, "extra_attr")
        assert isinstance(arr, subclass_uarray)

    def test_init_from_unyt_array(self):
        """
        Check initializing from a unyt_array.
        """
        arr = subclass_uarray(
            u.unyt_array(np.ones(5), units=u.Mpc),
            extra_attr=False,
        )
        assert hasattr(arr, "extra_attr")
        assert isinstance(arr, subclass_uarray)

    def test_init_from_list_of_unyt_arrays(self):
        """
        Check initializing from a list of unyt_array's.

        Note that unyt won't recurse deeper than one level on inputs, so we don't test
        deeper than one level of lists.
        """
        arr = subclass_uarray(
            [u.unyt_array(1, units=u.Mpc) for _ in range(5)],
            extra_attr=True,
        )
        assert hasattr(arr, "extra_attr")
        assert isinstance(arr, subclass_uarray)

    def test_init_from_list_of_subclass_arrays(self):
        """
        Check initializing from a list of subclass_uarray's.

        Note that unyt won't recurse deeper than one level on inputs, so we don't test
        deeper than one level of lists.
        """
        arr = subclass_uarray(
            [
                subclass_uarray(
                    [1],
                    units=u.Mpc,
                    extra_attr=True,
                )
                for _ in range(5)
            ]
        )
        assert isinstance(arr, subclass_uarray)
        assert hasattr(arr, "extra_attr") and arr.extra_attr is True

    def test_expected_init_failures(self):
        # we refuse mixed extra_attr:
        with pytest.raises(ExtraAttributeError):
            subclass_uarray(
                [
                    subclass_uquantity(1, u.Mpc, extra_attr=True),
                    subclass_uquantity(1, u.Mpc, extra_attr=False),
                ]
            )
        for cls, inp in ((subclass_uarray, [1]), (subclass_uquantity, 1)):
            # we refuse overriding an input subclass_uarray's information:
            with pytest.raises(ValueError):
                cls(
                    cls(inp, units=u.Mpc, extra_attr=False),
                    units=u.Mpc,
                    extra_attr=True,
                )
            # unless it matches, that's fine:
            cls(
                cls(
                    inp,
                    units=u.Mpc,
                    extra_attr=True,
                ),
                units=u.Mpc,
                extra_attr=True,
            )


class TestNumpyFunctions:
    """
    Check that numpy functions recognize our subclasses as input and handle them
    correctly.
    """

    def test_explicitly_handled_funcs(self):
        """
        Make sure we at least handle everything that unyt does, and anything that
        'just worked' for unyt but that we need to handle by hand.

        We don't try to be exhaustive here, but at give some basic input to every function
        that we expect to be able to take subclass input. We then use our helpers defined
        above to convert the inputs to unyt equivalents and call the numpy function on
        both subclass and unyt input. Then we use our helpers to check the results for
        consistency. For instance, if with unyt input we got back a unyt_array, we
        should expect a subclass_uarray.

        We are not currently explicitly testing that the results of any specific function
        are numerically what we expected them to be (seems like overkill), nor that the
        extra_attr's make sense given the input. The latter would be a useful addition,
        but I can't think of a sensible way to implement this besides writing in the
        expectation for every output value of every function by hand.

        As long as no functions outright crash, the test will report the list of functions
        that we should have covered that we didn't cover in tests, and/or the list of
        functions whose output values were not what we expected based on running them with
        unyt input. Otherwise we just get a stack trace of the first function that
        crashed.
        """
        from unyt._array_functions import _HANDLED_FUNCTIONS
        from unyt.tests.test_array_functions import NOOP_FUNCTIONS

        functions_to_check = {
            # FUNCTIONS UNYT HANDLES EXPLICITLY:
            "array2string": (sub_arr(np.arange(3)),),
            "dot": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "vdot": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "inner": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "outer": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "kron": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "histogram_bin_edges": (sub_arr(np.arange(3)),),
            "linalg.inv": (sub_arr(np.eye(3)),),
            "linalg.tensorinv": (sub_arr(np.eye(9).reshape((3, 3, 3, 3))),),
            "linalg.pinv": (sub_arr(np.eye(3)),),
            "linalg.svd": (sub_arr(np.eye(3)),),
            "histogram": (sub_arr(np.arange(3)),),
            "histogram2d": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "histogramdd": (sub_arr(np.arange(3)).reshape((1, 3)),),
            "concatenate": (sub_arr(np.eye(3)),),
            "cross": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "intersect1d": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "union1d": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "linalg.norm": (sub_arr(np.arange(3)),),
            "vstack": (sub_arr(np.arange(3)),),
            "hstack": (sub_arr(np.arange(3)),),
            "dstack": (sub_arr(np.arange(3)),),
            "column_stack": (sub_arr(np.arange(3)),),
            "stack": (sub_arr(np.arange(3)),),
            "around": (sub_arr(np.arange(3)),),
            "block": ([[sub_arr(np.arange(3))], [sub_arr(np.arange(3))]],),
            "fft.fft": (sub_arr(np.arange(3)),),
            "fft.fft2": (sub_arr(np.eye(3)),),
            "fft.fftn": (sub_arr(np.arange(3)),),
            "fft.hfft": (sub_arr(np.arange(3)),),
            "fft.rfft": (sub_arr(np.arange(3)),),
            "fft.rfft2": (sub_arr(np.eye(3)),),
            "fft.rfftn": (sub_arr(np.arange(3)),),
            "fft.ifft": (sub_arr(np.arange(3)),),
            "fft.ifft2": (sub_arr(np.eye(3)),),
            "fft.ifftn": (sub_arr(np.arange(3)),),
            "fft.ihfft": (sub_arr(np.arange(3)),),
            "fft.irfft": (sub_arr(np.arange(3)),),
            "fft.irfft2": (sub_arr(np.eye(3)),),
            "fft.irfftn": (sub_arr(np.arange(3)),),
            "fft.fftshift": (sub_arr(np.arange(3)),),
            "fft.ifftshift": (sub_arr(np.arange(3)),),
            "sort_complex": (sub_arr(np.arange(3)),),
            "isclose": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "allclose": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "array_equal": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "array_equiv": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "linspace": (sub_q(1), sub_q(2)),
            "logspace": (
                sub_q(1, unit=u.dimensionless),
                sub_q(2, unit=u.dimensionless),
            ),
            "geomspace": (sub_q(1), sub_q(1)),
            "copyto": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "prod": (sub_arr(np.arange(3)),),
            "var": (sub_arr(np.arange(3)),),
            "trace": (sub_arr(np.eye(3)),),
            "percentile": (sub_arr(np.arange(3)), 30),
            "quantile": (sub_arr(np.arange(3)), 0.3),
            "nanpercentile": (sub_arr(np.arange(3)), 30),
            "nanquantile": (sub_arr(np.arange(3)), 0.3),
            "linalg.det": (sub_arr(np.eye(3)),),
            "diff": (sub_arr(np.arange(3)),),
            "ediff1d": (sub_arr(np.arange(3)),),
            "ptp": (sub_arr(np.arange(3)),),
            "cumprod": (sub_arr(np.arange(3)),),
            "pad": (sub_arr(np.arange(3)), 3),
            "choose": (np.arange(3), sub_arr(np.eye(3))),
            "insert": (sub_arr(np.arange(3)), 1, sub_q(1)),
            "linalg.lstsq": (sub_arr(np.eye(3)), sub_arr(np.eye(3))),
            "linalg.solve": (sub_arr(np.eye(3)), sub_arr(np.eye(3))),
            "linalg.tensorsolve": (
                sub_arr(np.eye(24).reshape((6, 4, 2, 3, 4))),
                sub_arr(np.ones((6, 4))),
            ),
            "linalg.eig": (sub_arr(np.eye(3)),),
            "linalg.eigh": (sub_arr(np.eye(3)),),
            "linalg.eigvals": (sub_arr(np.eye(3)),),
            "linalg.eigvalsh": (sub_arr(np.eye(3)),),
            "savetxt": (savetxt_file, sub_arr(np.arange(3))),
            "fill_diagonal": (
                sub_arr(np.eye(3)),
                sub_arr(np.arange(3)),
            ),
            "apply_over_axes": (lambda x, axis: x, sub_arr(np.eye(3)), (0, 1)),
            "isin": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "place": (
                sub_arr(np.arange(3)),
                np.arange(3) > 0,
                sub_arr(np.arange(3)),
            ),
            "put": (
                sub_arr(np.arange(3)),
                np.arange(3),
                sub_arr(np.arange(3)),
            ),
            "put_along_axis": (
                sub_arr(np.arange(3)),
                np.arange(3),
                sub_arr(np.arange(3)),
                0,
            ),
            "putmask": (
                sub_arr(np.arange(3)),
                np.arange(3),
                sub_arr(np.arange(3)),
            ),
            "searchsorted": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "select": (
                [np.arange(3) < 1, np.arange(3) > 1],
                [sub_arr(np.arange(3)), sub_arr(np.arange(3))],
                sub_q(1),
            ),
            "setdiff1d": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3, 6)),
            ),
            "sinc": (sub_arr(np.arange(3)),),
            "clip": (sub_arr(np.arange(3)), sub_q(1), sub_q(2)),
            "where": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "triu": (sub_arr(np.ones((3, 3))),),
            "tril": (sub_arr(np.ones((3, 3))),),
            "einsum": ("ii->i", sub_arr(np.eye(3))),
            "convolve": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "correlate": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "unwrap": (sub_arr(np.arange(3)),),
            "interp": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "array_repr": (sub_arr(np.arange(3)),),
            trapezoid_fname: (sub_arr(np.arange(3)),),
            "take": (sub_arr(np.arange(3)), np.arange(3)),
            # FUNCTIONS THAT UNYT DOESN'T HANDLE EXPLICITLY (THEY "JUST WORK"):
            "all": (sub_arr(np.arange(3)),),
            "amax": (sub_arr(np.arange(3)),),  # implemented via max
            "amin": (sub_arr(np.arange(3)),),  # implemented via min
            "angle": (sub_q(complex(1, 1)),),
            "any": (sub_arr(np.arange(3)),),
            "append": (sub_arr(np.arange(3)), sub_q(1)),
            "apply_along_axis": (lambda x: x, 0, sub_arr(np.eye(3))),
            "argmax": (sub_arr(np.arange(3)),),  # implemented via max
            "argmin": (sub_arr(np.arange(3)),),  # implemented via min
            "argpartition": (
                sub_arr(np.arange(3)),
                1,
            ),  # implemented via partition
            "argsort": (sub_arr(np.arange(3)),),  # implemented via sort
            "argwhere": (sub_arr(np.arange(3)),),
            "array_str": (sub_arr(np.arange(3)),),
            "atleast_1d": (sub_arr(np.arange(3)),),
            "atleast_2d": (sub_arr(np.arange(3)),),
            "atleast_3d": (sub_arr(np.arange(3)),),
            "average": (sub_arr(np.arange(3)),),
            "can_cast": (sub_arr(np.arange(3)), np.float64),
            "common_type": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "result_type": (sub_arr(np.ones(3)), sub_arr(np.ones(3))),
            "iscomplex": (sub_arr(np.arange(3)),),
            "iscomplexobj": (sub_arr(np.arange(3)),),
            "isreal": (sub_arr(np.arange(3)),),
            "isrealobj": (sub_arr(np.arange(3)),),
            "nan_to_num": (sub_arr(np.arange(3)),),
            "nanargmax": (sub_arr(np.arange(3)),),  # implemented via max
            "nanargmin": (sub_arr(np.arange(3)),),  # implemented via min
            "nanmax": (sub_arr(np.arange(3)),),  # implemented via max
            "nanmean": (sub_arr(np.arange(3)),),  # implemented via mean
            "nanmedian": (sub_arr(np.arange(3)),),  # implemented via median
            "nanmin": (sub_arr(np.arange(3)),),  # implemented via min
            "trim_zeros": (sub_arr(np.arange(3)),),
            "max": (sub_arr(np.arange(3)),),
            "mean": (sub_arr(np.arange(3)),),
            "median": (sub_arr(np.arange(3)),),
            "min": (sub_arr(np.arange(3)),),
            "ndim": (sub_arr(np.arange(3)),),
            "shape": (sub_arr(np.arange(3)),),
            "size": (sub_arr(np.arange(3)),),
            "sort": (sub_arr(np.arange(3)),),
            "sum": (sub_arr(np.arange(3)),),
            "repeat": (sub_arr(np.arange(3)), 2),
            "tile": (sub_arr(np.arange(3)), 2),
            "shares_memory": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "nonzero": (sub_arr(np.arange(3)),),
            "count_nonzero": (sub_arr(np.arange(3)),),
            "flatnonzero": (sub_arr(np.arange(3)),),
            "isneginf": (sub_arr(np.arange(3)),),
            "isposinf": (sub_arr(np.arange(3)),),
            "empty_like": (sub_arr(np.arange(3)),),
            "full_like": (sub_arr(np.arange(3)), sub_q(1)),
            "ones_like": (sub_arr(np.arange(3)),),
            "zeros_like": (sub_arr(np.arange(3)),),
            "copy": (sub_arr(np.arange(3)),),
            "meshgrid": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "transpose": (sub_arr(np.eye(3)),),
            "reshape": (sub_arr(np.arange(3)), (3,)),
            "resize": (sub_arr(np.arange(3)), 6),
            "roll": (sub_arr(np.arange(3)), 1),
            "rollaxis": (sub_arr(np.arange(3)), 0),
            "rot90": (sub_arr(np.eye(3)),),
            "expand_dims": (sub_arr(np.arange(3)), 0),
            "squeeze": (sub_arr(np.arange(3)),),
            "flip": (sub_arr(np.eye(3)),),
            "fliplr": (sub_arr(np.eye(3)),),
            "flipud": (sub_arr(np.eye(3)),),
            "delete": (sub_arr(np.arange(3)), 0),
            "partition": (sub_arr(np.arange(3)), 1),
            "broadcast_to": (sub_arr(np.arange(3)), 3),
            "broadcast_arrays": (sub_arr(np.arange(3)),),
            "split": (sub_arr(np.arange(3)), 1),
            "array_split": (sub_arr(np.arange(3)), 1),
            "dsplit": (sub_arr(np.arange(27)).reshape(3, 3, 3), 1),
            "hsplit": (sub_arr(np.arange(3)), 1),
            "vsplit": (sub_arr(np.eye(3)), 1),
            "swapaxes": (sub_arr(np.eye(3)), 0, 1),
            "moveaxis": (sub_arr(np.eye(3)), 0, 1),
            "nansum": (sub_arr(np.arange(3)),),  # implemented via sum
            "std": (sub_arr(np.arange(3)),),
            "nanstd": (sub_arr(np.arange(3)),),
            "nanvar": (sub_arr(np.arange(3)),),
            "nanprod": (sub_arr(np.arange(3)),),
            "diag": (sub_arr(np.eye(3)),),
            "diag_indices_from": (sub_arr(np.eye(3)),),
            "diagflat": (sub_arr(np.eye(3)),),
            "diagonal": (sub_arr(np.eye(3)),),
            "ravel": (sub_arr(np.arange(3)),),
            "ravel_multi_index": (np.eye(2, dtype=int), (2, 2)),
            "unravel_index": (np.arange(3), (3,)),
            "fix": (sub_arr(np.arange(3)),),
            "round": (sub_arr(np.arange(3)),),  # implemented via around
            "may_share_memory": (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            ),
            "linalg.matrix_power": (sub_arr(np.eye(3)), 2),
            "linalg.cholesky": (sub_arr(np.eye(3)),),
            "linalg.multi_dot": ((sub_arr(np.eye(3)), sub_arr(np.eye(3))),),
            "linalg.matrix_rank": (sub_arr(np.eye(3)),),
            "linalg.qr": (sub_arr(np.eye(3)),),
            "linalg.slogdet": (sub_arr(np.eye(3)),),
            "linalg.cond": (sub_arr(np.eye(3)),),
            "tensordot": (sub_arr(np.eye(3)), sub_arr(np.eye(3))),
            "gradient": (sub_arr(np.arange(3)),),
            "cumsum": (sub_arr(np.arange(3)),),
            "nancumsum": (sub_arr(np.arange(3)),),
            "nancumprod": (sub_arr(np.arange(3)),),
            "bincount": (sub_arr(np.arange(3)),),
            "unique": (sub_arr(np.arange(3)),),
            "min_scalar_type": (sub_arr(np.arange(3)),),
            "extract": (0, sub_arr(np.arange(3))),
            "setxor1d": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "lexsort": (sub_arr(np.arange(3)),),
            "digitize": (sub_arr(np.arange(3)), sub_arr(np.arange(3))),
            "tril_indices_from": (sub_arr(np.eye(3)),),
            "triu_indices_from": (sub_arr(np.eye(3)),),
            "imag": (sub_arr(np.arange(3)),),
            "real": (sub_arr(np.arange(3)),),
            "real_if_close": (sub_arr(np.arange(3)),),
            "einsum_path": (
                "ij,jk->ik",
                sub_arr(np.eye(3)),
                sub_arr(np.eye(3)),
            ),
            "cov": (sub_arr(np.arange(3)),),
            "corrcoef": (sub_arr(np.arange(3)),),
            "compress": (np.zeros(3), sub_arr(np.arange(3))),
            "take_along_axis": (
                sub_arr(np.arange(3)),
                np.ones(3, dtype=int),
                0,
            ),
        }
        if NUMPY_VERSION >= Version("2.0.0dev0"):
            functions_to_check["linalg.cross"] = (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            )
            functions_to_check["linalg.diagonal"] = (sub_arr(np.eye(3)),)
            functions_to_check["linalg.matmul"] = (
                sub_arr(np.eye(3)),
                sub_arr(np.eye(3)),
            )
            functions_to_check["linalg.matrix_norm"] = (sub_arr(np.eye(3)),)
            functions_to_check["linalg.matrix_transpose"] = (sub_arr(np.eye(3)),)
            functions_to_check["linalg.tensordot"] = (
                sub_arr(np.eye(3)),
                sub_arr(np.eye(3)),
            )
            functions_to_check["linalg.outer"] = (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            )
            functions_to_check["linalg.svdvals"] = (sub_arr(np.eye(3)),)
            functions_to_check["linalg.trace"] = (sub_arr(np.eye(3)),)
            functions_to_check["linalg.vecdot"] = (
                sub_arr(np.arange(3)),
                sub_arr(np.arange(3)),
            )
            functions_to_check["linalg.vector_norm"] = (sub_arr(np.arange(3)),)
            functions_to_check["astype"] = (sub_arr(np.arange(3)), float)
            functions_to_check["matrix_transpose"] = (sub_arr(np.eye(3)),)
            functions_to_check["unique_all"] = (sub_arr(np.arange(3)),)
            functions_to_check["unique_counts"] = (sub_arr(np.arange(3)),)
            functions_to_check["unique_inverse"] = (sub_arr(np.arange(3)),)
            functions_to_check["unique_values"] = (sub_arr(np.arange(3)),)
            functions_to_check["cumulative_sum"] = (sub_arr(np.arange(3)),)
            functions_to_check["cumulative_prod"] = (sub_arr(np.arange(3)),)
            functions_to_check["unstack"] = (sub_arr(np.arange(3)),)
        else:
            functions_to_check["asfarray"] = (sub_arr(np.arange(3)),)

        if NUMPY_VERSION < Version("2.4.0.dev0"):
            functions_to_check["in1d"] = (sub_arr(np.arange(3)), sub_arr(np.arange(3)))

        functions_checked = []
        bad_funcs = {}
        for fname, args in functions_to_check.items():
            ua_args = []
            for arg in args:
                ua_args.append(arg_to_ua(arg))
            func = getfunc(fname)
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action="ignore",
                        category=FutureWarning,
                    )
                    if "savetxt" in fname:
                        warnings.filterwarnings(
                            action="ignore",
                            category=UserWarning,
                            message="numpy.savetxt does not preserve units",
                        )
                    try:
                        ua_result = func(*ua_args)
                    except:
                        print(f"Crashed in {fname} with unyt input.")
                        raise
            except u.exceptions.UnytError:
                raises_unyt_error = True
            else:
                raises_unyt_error = False
            if "savetxt" in fname and os.path.isfile(savetxt_file):
                os.remove(savetxt_file)
            functions_checked.append(func)
            if raises_unyt_error:
                with pytest.raises(u.exceptions.UnytError):
                    result = func(*args)
                continue
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=FutureWarning,
                )
                if "savetxt" in fname:
                    warnings.filterwarnings(
                        action="ignore",
                        category=UserWarning,
                        message="numpy.savetxt does not preserve units or extra_attr",
                    )
                try:
                    result = func(*args)
                except:
                    print(f"Crashed in {fname} with subclass input.")
                    raise
            if fname.split(".")[-1] in (
                "fill_diagonal",
                "copyto",
                "place",
                "put",
                "put_along_axis",
                "putmask",
            ):
                # treat inplace modified values for relevant functions as result
                result = args[0]
                ua_result = ua_args[0]
            if "savetxt" in fname and os.path.isfile(savetxt_file):
                os.remove(savetxt_file)
            ignore_values = fname in {"empty_like"}  # empty_like has arbitrary data
            try:
                check_result(result, ua_result, ignore_values=ignore_values)
            except AssertionError:
                bad_funcs["np." + fname] = result, ua_result
        if len(bad_funcs) > 0:
            raise AssertionError(
                "Some functions did not return expected types "
                "(obtained, obtained with unyt input): " + str(bad_funcs)
            )
        unchecked_functions = [
            f
            for f in set(_HANDLED_FUNCTIONS) | NOOP_FUNCTIONS
            if f not in functions_checked
        ]
        try:
            assert len(unchecked_functions) == 0
        except AssertionError:
            raise AssertionError(
                "Did not check functions",
                [
                    (".".join((f.__module__, f.__name__)).replace("numpy", "np"))
                    for f in unchecked_functions
                ],
            )

    @pytest.mark.parametrize(
        "func_args",
        (
            (
                np.histogram,
                (
                    subclass_uarray(
                        [1, 2, 3],
                        u.m,
                        extra_attr=True,
                    ),
                ),
            ),
            (
                np.histogram2d,
                (
                    subclass_uarray(
                        [1, 2, 3],
                        u.m,
                        extra_attr=True,
                    ),
                    subclass_uarray(
                        [1, 2, 3],
                        u.K,
                        extra_attr=True,
                    ),
                ),
            ),
            (
                np.histogramdd,
                (
                    [
                        subclass_uarray(
                            [1, 2, 3],
                            u.m,
                            extra_attr=True,
                        ),
                        subclass_uarray(
                            [1, 2, 3],
                            u.K,
                            extra_attr=True,
                        ),
                        subclass_uarray(
                            [1, 2, 3],
                            u.kg,
                            extra_attr=True,
                        ),
                    ],
                ),
            ),
        ),
    )
    @pytest.mark.parametrize(
        "weights",
        (
            None,
            subclass_uarray([1, 2, 3], u.s, extra_attr=True),
            np.array([1, 2, 3]),
        ),
    )
    @pytest.mark.parametrize("bins_type", ("int", "sub_arr"))
    @pytest.mark.parametrize("density", (None, True))
    def test_histograms(self, func_args, weights, bins_type, density):
        """
        Test that histograms give sensible output.

        Histograms are tricky with possible density and weights arguments, and the way
        that attributes need validation and propagation between the bins and values.
        They are also commonly used. They therefore need a bespoke test.
        """
        func, args = func_args
        bins = {
            "int": 10,
            "sub_arr": [
                subclass_uarray(
                    np.linspace(0, 5, 11),
                    u.kpc,
                    extra_attr=True,
                ),
                subclass_uarray(
                    np.linspace(0, 5, 11),
                    u.K,
                    extra_attr=True,
                ),
                subclass_uarray(
                    np.linspace(0, 5, 11),
                    u.Msun,
                    extra_attr=True,
                ),
            ],
        }[bins_type]
        bins = (
            bins[
                {
                    np.histogram: np.s_[0],
                    np.histogram2d: np.s_[:2],
                    np.histogramdd: np.s_[:],
                }[func]
            ]
            if bins_type == "sub_arr"
            else bins
        )
        result = func(*args, bins=bins, density=density, weights=weights)
        ua_args = tuple(
            (
                to_ua(arg)
                if not isinstance(arg, tuple)
                else tuple(to_ua(item) for item in arg)
            )
            for arg in args
        )
        ua_bins = (
            to_ua(bins)
            if not isinstance(bins, tuple)
            else tuple(to_ua(item) for item in bins)
        )
        ua_result = func(
            *ua_args, bins=ua_bins, density=density, weights=to_ua(weights)
        )
        if isinstance(ua_result, tuple):
            assert isinstance(result, tuple)
            assert len(result) == len(ua_result)
            for r, ua_r in zip(result, ua_result):
                check_result(r, ua_r)
        else:
            check_result(result, ua_result)
        if not density and not isinstance(weights, subclass_uarray):
            assert not isinstance(result[0], subclass_uarray)
        else:
            assert result[0].extra_attr is True
        if density and not isinstance(weights, subclass_uarray):
            pass
        elif density and isinstance(weights, subclass_uarray):
            assert result[0].extra_attr is True
        elif not density and isinstance(weights, subclass_uarray):
            assert result[0].extra_attr is True
        ret_bins = {
            np.histogram: [result[1]],
            np.histogram2d: result[1:],
            np.histogramdd: result[1],
        }[func]
        for b in ret_bins:
            assert b.extra_attr is True

    def test_getitem(self):
        """
        Make sure that we don't degrade to an ndarray on slicing.
        """
        assert isinstance(sub_arr(np.arange(3))[0], subclass_uquantity)

    def test_reshape_to_scalar(self):
        """
        Make sure that we convert to a subclass_uquantity when we reshape to a scalar.
        """
        assert isinstance(sub_arr(np.ones(1)).reshape(()), subclass_uquantity)

    def test_iter(self):
        """
        Make sure that we get subclass_uquantity's when iterating over a subclass_uarray.
        """
        for sub_q in sub_arr(np.arange(3)):
            assert isinstance(sub_q, subclass_uquantity)

    def test_dot(self):
        """
        Make sure that we get a subclass_uarray when we use array attribute dot.
        """
        res = sub_arr(np.arange(3)).dot(sub_arr(np.arange(3)))
        assert isinstance(res, subclass_uquantity)
        assert res.extra_attr is True


class TestSubclassQuantity:
    """
    Test that the subclass_uquantity class works as desired, mostly around issues
    converting back and forth with subclass_uarray.
    """

    @pytest.mark.parametrize(
        "func, args",
        [
            ("astype", (float,)),
            ("in_units", (u.m,)),
            ("byteswap", ()),
            ("compress", ([True],)),
            ("flatten", ()),
            ("ravel", ()),
            ("repeat", (1,)),
            ("reshape", (1,)),
            ("take", ([0],)),
            ("transpose", ()),
            ("view", ()),
        ],
    )
    def test_propagation_func(self, func, args):
        """
        Test that functions that are supposed to propagate our attribute do so.
        """
        sub_q = subclass_uquantity(
            1,
            u.m,
            extra_attr=True,
        )
        res = getattr(sub_q, func)(*args)
        assert res.extra_attr is True

    def test_round(self):
        """
        Test that attributes propagate through the round builtin.
        """
        sub_q = subclass_uquantity(
            1.03,
            u.m,
            extra_attr=True,
        )
        res = round(sub_q)
        assert res.value == 1.0
        assert res.extra_attr is True

    def test_scalar_return_func(self):
        """
        Make sure that default-wrapped functions that take a subclass_uarray and return a
        scalar convert to a subclass_uquantity.
        """
        sub_arr = subclass_uarray(
            np.arange(3),
            u.m,
            extra_attr=True,
        )
        res = np.min(sub_arr)
        assert isinstance(res, subclass_uquantity)

    @pytest.mark.parametrize("prop", ["T", "ua", "unit_array"])
    def test_propagation_props(self, prop):
        """
        Test that properties propagate our attributes as intended.
        """
        sub_q = subclass_uquantity(
            1,
            u.m,
            extra_attr=True,
        )
        res = getattr(sub_q, prop)
        assert res.extra_attr is True

    def test_multiply_quantities(self):
        """
        Test multiplying two quantities.
        """
        sub_q = subclass_uquantity(
            2,
            u.m,
            extra_attr=True,
        )
        multiplied = sub_q * sub_q
        assert type(multiplied) is subclass_uquantity
        assert multiplied.extra_attr is True
        assert multiplied.to_value(u.m**2) == 4


class TestCosmoArrayCopy:
    """
    Tests of explicit (deep)copying of subclass_uarray.
    """

    def test_copy(self):
        """
        Check that when we copy a subclass_uarray it preserves its values and attributes.
        """
        units = u.Mpc
        arr = subclass_uarray(
            u.unyt_array(np.ones(5), units=units),
            extra_attr=True,
        )
        copy_arr = copy(arr)
        assert np.allclose(arr.to_value(units), copy_arr.to_value(units))
        assert arr.units == copy_arr.units
        assert arr.extra_attr == copy_arr.extra_attr

    def test_deepcopy(self):
        """
        Check that when we deepcopy a subclass_uarray it preserves its values and
        attribute
        """
        units = u.Mpc
        arr = subclass_uarray(
            u.unyt_array(np.ones(5), units=units),
            extra_attr=True,
        )
        copy_arr = deepcopy(arr)
        assert np.allclose(arr.to_value(units), copy_arr.to_value(units))
        assert arr.units == copy_arr.units
        assert arr.extra_attr == copy_arr.extra_attr

    def test_to_cgs(self):
        """
        Check that using to_cgs properly preserves attributes.
        """
        units = u.Mpc
        arr = subclass_uarray(
            u.unyt_array(np.ones(5), units=units),
            extra_attr=True,
        )
        cgs_arr = arr.in_cgs()
        assert np.allclose(arr.to_value(u.cm), cgs_arr.to_value(u.cm))
        assert cgs_arr.units == u.cm
        assert cgs_arr.extra_attr == arr.extra_attr


class TestMultiplicationByUnyt:
    @pytest.mark.parametrize(
        "sub_arr",
        [
            subclass_uarray(
                np.ones(3),
                u.Mpc,
                extra_attr=True,
            ),
            subclass_uquantity(
                np.ones(1),
                u.Mpc,
                extra_attr=True,
            ),
        ],
    )
    def test_multiplication_by_unyt(self, sub_arr):
        """
        We desire consistent behaviour for example for
        `subclass_uarray(...) * (1 * u.Mpc)` as for
        `subclass_uarray(...) * u.Mpc`.
        """

        lmultiplied_by_quantity = sub_arr * (
            1 * u.Mpc
        )  # parentheses very important here
        lmultiplied_by_unyt = sub_arr * u.Mpc
        assert isinstance(lmultiplied_by_quantity, subclass_uarray)
        assert isinstance(lmultiplied_by_unyt, subclass_uarray)
        assert lmultiplied_by_unyt.extra_attr == lmultiplied_by_quantity.extra_attr
        assert np.allclose(
            lmultiplied_by_unyt.to_value(lmultiplied_by_quantity.units),
            lmultiplied_by_quantity.to_value(lmultiplied_by_quantity.units),
        )

        ldivided_by_quantity = sub_arr / (1 * u.Mpc)  # parentheses very important here
        ldivided_by_unyt = sub_arr / u.Mpc
        assert isinstance(ldivided_by_quantity, subclass_uarray)
        assert isinstance(ldivided_by_unyt, subclass_uarray)
        assert ldivided_by_unyt.extra_attr == ldivided_by_quantity.extra_attr
        assert np.allclose(
            ldivided_by_unyt.to_value(ldivided_by_quantity.units),
            ldivided_by_quantity.to_value(ldivided_by_quantity.units),
        )

        rmultiplied_by_quantity = (
            1 * u.Mpc
        ) * sub_arr  # parentheses very important here
        assert rmultiplied_by_quantity.extra_attr
        rmultiplied_by_unyt = u.Mpc * sub_arr
        assert isinstance(rmultiplied_by_quantity, subclass_uarray)
        assert isinstance(rmultiplied_by_unyt, subclass_uarray)
        assert rmultiplied_by_unyt.extra_attr == rmultiplied_by_quantity.extra_attr
        assert np.allclose(
            rmultiplied_by_unyt.to_value(rmultiplied_by_quantity.units),
            rmultiplied_by_quantity.to_value(rmultiplied_by_quantity.units),
        )

        rdivided_by_quantity = (1 * u.Mpc) / sub_arr  # parentheses very important here
        rdivided_by_unyt = u.Mpc / sub_arr
        assert isinstance(rdivided_by_quantity, subclass_uarray)
        assert isinstance(rdivided_by_unyt, subclass_uarray)
        assert rdivided_by_unyt.extra_attr == rdivided_by_quantity.extra_attr
        assert np.allclose(
            rdivided_by_unyt.to_value(rdivided_by_quantity.units),
            rdivided_by_quantity.to_value(rdivided_by_quantity.units),
        )
